# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

import math
import numpy as np
from NEURAL_py_EEG import utils
from NEURAL_py_EEG import NEURAL_parameters

# noinspection SpellCheckingInspection
def gen_spectrum(x, Fs, param_st, SCALE_PSD=0):
    """
    gen_spectrum: specturm using either:
                            1. periodogram
                            2. Welch's PSD
                            3. robust Welch PSD
    Syntax: pxx,itotal_bandpass,f_scale,Nfreq,fp = gen_spectrum(x,Fs,param_st,SCALE_PSD)
    Inputs:
        x             - input signal
        Fs            - sampling frequency (in Hz)
        param_st      - parameter structure (with window length, window type, overlap,
                        frequency bands, total frequency band, and method)
        SCALE_PSD     - scale PSD by factor of 2 (expect DC and Nyquist)? (0=no [default]
                        or 1=yes)
    Outputs:
        pxx              - spectral estimate (e.g. power spectral density)
        itotal_bandpass  - indices for total frequency band (defined in input args.)
        f_scale          - frequency scaling factor
        Nfreq            - total length of pxx (including -ve frequencies)
        fp

    Example:
        import utils
        import spectral_features
        import matplotlib.pyplot as plt
        import numpy as np

        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values

        param_st = dict()
        param_st['method'] = 'robust-PSD'
        param_st['L_window'] = 2
        param_st['window_type'] = 'hamm'
        param_st['overlap'] = 50
        param_st['freq_bands'] = [[0.5, 4], [4, 7], [7, 13], [13, 30]]
        param_st['total_freq_bands'] = [0.5, 30]

        [pxx,itotal_bandpass,~,~,fp]=spectral_features.gen_spectrum(x,Fs,param_st)

        plt.figure(1)
        plt.plot(fp, 10*np.log10(pxx))
        plt.plot([fp[itotal_bandpass[0]], fp[itotal_bandpass[0]]], [30, -30], 'k')
        plt.plot([fp[itotal_bandpass[-1]],  fp[itotal_bandpass[-1]]], [30, -30], 'k')
        plt.show()


    """
    L_window = param_st["L_window"]
    window_type = param_st["window_type"]
    overlap = param_st["overlap"]
    spec_method = param_st["method"].lower()

    # remove nans
    x = np.delete(x, np.where(np.isnan(x))[0])

    if spec_method == "bartlett-psd":
        # ---------------------------------------------------------------------
        # Bartlett PSD: same as Welch with 0 % overlap and rectangular window
        # ---------------------------------------------------------------------
        window_type = "rect"
        overlap = 0
        spec_method = "psd"

    if spec_method == "psd":
        # ---------------------------------------------------------------------
        # Welch PSD
        # ---------------------------------------------------------------------
        S_stft, Nfreq, f_scale, win_epoch = gen_STFT(
            x, L_window, window_type, overlap, Fs
        )
        pxx = np.nanmean(S_stft, 0)
        # N = len(pxx)

        # Normalise (so similar to pwelch)
        E_win = np.sum(np.abs(win_epoch) ** 2) / Nfreq
        pxx = pxx / (Nfreq * E_win * Fs)

    elif spec_method == "robust-psd":
        # ---------------------------------------------------------------------
        # Welch PSD with median instead of mean
        # ---------------------------------------------------------------------
        S_stft, Nfreq, f_scale, win_epoch = gen_STFT(
            x, L_window, window_type, overlap, Fs
        )
        pxx = np.nanmedian(S_stft, 0)
        # N = len(pxx) # Need to verify this

        # Normalise (so similar to pwelch)
        E_win = np.sum(np.abs(win_epoch) ** 2) / Nfreq
        pxx = pxx / (Nfreq * E_win * Fs)
    elif spec_method == "periodogram":
        # ---------------------------------------------------------------------
        # Periodogram
        # ---------------------------------------------------------------------
        X = np.abs(np.fft.fft(x)) ** 2

        # +ve frequencies only
        N = len(X)
        Nh = np.floor(N / 2).astype(int)
        X = X[range(Nh + 1)]
        Nfreq = N

        pxx = X / (Fs * N)

    else:
        raise ValueError(
            "Unknown spectral method - check spelling of: %s" % spec_method
        )

    if SCALE_PSD:
        pscale = np.ones(len(pxx)) + 1
        if Nfreq % 2 == 1:
            pscale[0] = 1
        else:
            pscale[0] = 1
            pscale[-1] = 1
        pxx = pxx * pscale

    N = len(pxx)
    f_scale = Nfreq / Fs
    fp = np.array(range(N)) / f_scale

    if "total_freq_bands" in param_st.keys():
        total_freq_bands = param_st["total_freq_bands"]

        # b) limit to frequency band of interest:
        itotal_bandpass = np.array(
            range(
                np.ceil(total_freq_bands[0] * f_scale).astype(int),
                np.floor(total_freq_bands[1] * f_scale).astype(int) + 1,
            )
        )
        itotal_bandpass = itotal_bandpass  # + 1
        itotal_bandpass[itotal_bandpass < 0] = 0
        itotal_bandpass[itotal_bandpass > N - 1] = N - 1
    else:
        itotal_bandpass = np.nan

    return pxx, itotal_bandpass, f_scale, Nfreq, fp


def gen_STFT(x, L_window, window_type, overlap, Fs, STFT_OR_SPEC=0):
    """
    gen_STFT: Short-time Fourier transform (or spectrogram)
    Syntax: S_stft,Nfreq,f_scale,win_epoch = spectral_features.gen_STFT(x,L_window,window_type,overlap,Fs)
    Inputs:
        x            - input signal
        L_window     - window length
        window_type  - window type
        overlap      - percentage overlap
        Fs           - sampling frequency (Hz)
        STFT_OR_SPEC - return short-time Fourier transform (STFT) or spectrogram
                   (0=spectrogram [default] and 1=STFT)
    Outputs:
        S_stft     - spectrogram
        Nfreq      - length of FFT
        f_scale    - frequency scaling factor
        win_epoch  - window

    Example:
        import utils
        import spectral_features
        import matplotlib.pyplot as plt

        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values

        L_window = 2
        window_type = 'hamm'
        overlap = 80

        S_stft, dum, dum, dum = spectral_features.gen_STFT(x, L_window, window_type, overlap, Fs)

        plt.figure(1)
        plt.imshow(S_stft, origin='lower')
        plt.colorbar()
        plt.show()
        import utils
        import spectral_features
        import matplotlib.pyplot as plt
    """

    L_hop, L_epoch, win_epoch = utils.gen_epoch_window(
        overlap, L_window, window_type, Fs, 1
    )

    N = len(x)
    N_epochs = math.floor((N - (L_epoch - L_hop)) / L_hop)
    if N_epochs < 1:
        N_epochs = 1

    nw = np.array(range(L_epoch))
    Nfreq = L_epoch

    # ---------------------------------------------------------------------
    # generate short - time FT on all data:
    # ---------------------------------------------------------------------
    K_stft = np.zeros([N_epochs, L_epoch])
    for k in range(N_epochs):
        nf = np.array((nw + k * L_hop) % N).astype(
            int
        )  # matlab used k-1 as this index using just k here
        # to start with no initial shift
        K_stft[k][:] = x[nf] * win_epoch

    f_scale = Nfreq / Fs

    S_stft = np.empty([N_epochs, L_epoch])
    if STFT_OR_SPEC:
        S_stft = S_stft.astype(complex)
        for k in range(N_epochs):
            S_stft[k][:] = np.fft.fft(K_stft[k][:], Nfreq)
    else:
        for k in range(N_epochs):
            S_stft[k][:] = np.abs(np.fft.fft(K_stft[k, :], Nfreq)) ** 2

    S_stft = S_stft[:, np.array(range(np.floor(Nfreq / 2).astype(int) + 1))]

    return S_stft, Nfreq, f_scale, win_epoch


# noinspection SpellCheckingInspection,DuplicatedCode
def spectral_power(x, Fs, feat_name, params_st):
    """

    :param x: Epoch data
    :param Fs: Sampling frequency
    :param feat_name: Name of feature 'spectral_power' or 'spectral_relative_power'
    :param params_st: Parameters
    :return: feature value
    """
    freq_bands = params_st["freq_bands"]
    params_st["method"] = "periodogram"
    pxx, itotal_bandpass, f_scale, N, fp = gen_spectrum(x, Fs, params_st, 1)
    pxx = pxx * Fs

    Nh = len(pxx)

    if feat_name == "spectral_relative_power":
        pxx_total = np.nansum(pxx[itotal_bandpass]) / N
    else:
        pxx_total = 1

    spec_power = np.empty(len(freq_bands))
    spec_power.fill(np.nan)

    ibandpass = []  # to suppress warning
    for p in range(len(freq_bands)):
        if p == 0:
            istart = np.ceil(freq_bands[p][0] * f_scale).astype(int) - 1
        else:
            istart = ibandpass[-1] - 1

        ibandpass = np.array(
            range(istart, np.floor(freq_bands[p][1] * f_scale).astype(int))
        )  # May need +1
        ibandpass = ibandpass + 1
        ibandpass[ibandpass < 0] = 0
        ibandpass[ibandpass > Nh - 1] = Nh - 1

        spec_power[p] = np.nansum(pxx[ibandpass] / (N * pxx_total))

    return spec_power


# noinspection SpellCheckingInspection
def spectral_flatness(x, Fs, feat_name, params_st):
    """
    ---------------------------------------------------------------------
     spectral flatness (Wiener entropy)
    ---------------------------------------------------------------------
    :param x: Epoch data
    :param Fs: Sampling frequency
    :param feat_name: Name of feature
    :param params_st: Parameters
    :return: feature value
    """
    pxx, dum, f_scale, dum, dum = gen_spectrum(x, Fs, params_st)

    # for each frequency band:
    freq_bands = params_st["freq_bands"]
    N_freq_bands = len(freq_bands)

    featx = np.empty(N_freq_bands)
    featx.fill(np.nan)
    N = len(pxx)
    ibandpass = []  # to suppress warning
    for p in range(N_freq_bands):
        if p == 0:
            istart = np.ceil(freq_bands[p][0] * f_scale).astype(int) - 1
        else:
            istart = ibandpass[-1] - 1
        ibandpass = np.array(
            range(istart, np.floor(freq_bands[p][1] * f_scale).astype(int))
        )
        ibandpass = ibandpass + 1
        ibandpass[ibandpass < 0] = 0
        ibandpass[ibandpass > N - 1] = N - 1

        featx[p] = np.exp(
            np.nanmean(np.log(pxx[ibandpass] + np.spacing(1)))
        ) / np.nanmean(pxx[ibandpass])

    return featx


def spectral_entropy(x, Fs, feat_name, params_st):
    """
    ---------------------------------------------------------------------
     spectral entropy (= Shannon entropy on normalised PSD)
    ---------------------------------------------------------------------
    :param x: Epoch data
    :param Fs: Sampling frequency
    :param feat_name: Name of feature
    :param params_st: Parameters
    :return: feature value
    """
    pxx, dum, f_scale, dum, dum = gen_spectrum(x, Fs, params_st)

    # for each frequency band:
    freq_bands = params_st["freq_bands"]
    N_freq_bands = len(freq_bands)

    featx = np.empty(N_freq_bands)
    featx.fill(np.nan)
    N = len(pxx)
    ibandpass = []  # to suppress warning
    for p in range(N_freq_bands):
        if p == 0:
            istart = np.ceil(freq_bands[p][0] * f_scale).astype(int) - 1
        else:
            istart = ibandpass[-1] - 1
        ibandpass = np.array(
            range(istart, np.floor(freq_bands[p][1] * f_scale).astype(int))
        )
        ibandpass = ibandpass + 1
        ibandpass[ibandpass < 0] = 0
        ibandpass[ibandpass > N - 1] = N - 1

        pr = pxx[ibandpass] / np.nansum(pxx[ibandpass])

        featx[p] = -np.nansum(pr * np.log(pr + np.spacing(1))) / np.log(len(pr))

    return featx


def spectral_diff(x, Fs, feat_name, params_st):
    """
    ---------------------------------------------------------------------
     spectral difference using the spectrogram
    ---------------------------------------------------------------------
    :param x: Epoch data
    :param Fs: Sampling frequency
    :param feat_name: Name of feature
    :param params_st: Parameters
    :return: feature value
    """
    # a) generate spectrogram
    S_stft, dum, f_scale, dum = gen_STFT(
        x, params_st["L_window"], params_st["window_type"], params_st["overlap"], Fs
    )

    N_epochs, M = S_stft.shape

    freq_bands = params_st["freq_bands"]
    N_freq_bands = len(freq_bands)

    featx = np.empty(N_freq_bands)
    featx.fill(np.nan)

    ibandpass = []  # to suppress warning
    for p in range(N_freq_bands):
        if p == 0:
            istart = np.ceil(freq_bands[p][0] * f_scale).astype(int) - 1
        else:
            istart = ibandpass[-1] - 1
        ibandpass = np.array(
            range(istart, np.floor(freq_bands[p][1] * f_scale).astype(int))
        )
        ibandpass = ibandpass + 1
        ibandpass[ibandpass < 0] = 0
        ibandpass[ibandpass > M - 1] = M - 1

        S_stft_band = S_stft[:, ibandpass] / np.nanmax(np.nanmax(S_stft[:, ibandpass]))

        spec_diff = np.zeros(N_epochs)
        for n in range(N_epochs - 1):
            v1 = S_stft_band[n, :]
            v2 = S_stft_band[n + 1, :]
            if all(np.isnan(v1)) or all(np.isnan(v2)):
                spec_diff[n] = np.nan
            else:
                spec_diff[n] = np.nanmean(np.abs(v1 - v2) ** 2)

        featx[p] = np.nanmedian(spec_diff)

    return featx


def spectral_edge_frequency(x, Fs, feat_name, params_st):
    """
    ---------------------------------------------------------------------
    spectral edge frequency
    ---------------------------------------------------------------------
    :param x: Epoch data
    :param Fs: Sampling frequency
    :param feat_name: Name of feature
    :param params_st: Parameters
    :return: feature value
    """
    pxx, itotal_bandpass, dum, dum, fp = gen_spectrum(x, Fs, params_st)

    # only within this frequency band:
    pxx[np.setdiff1d(range(len(pxx)), itotal_bandpass)] = 0

    pxx = pxx / np.nansum(pxx)

    # compute the cumulative density
    pxx_cum = np.nancumsum(pxx)

    # spectral edge frequency corresponds to the frequency at (nearest to)
    # the point on the freq axis where pyy_cum = 0.05
    idx = np.argmin(np.abs(pxx_cum - params_st["SEF"]), axis=0)
    return fp[idx]


def main_spectral(x, Fs, feat_name, params=None):
    """
    Syntax: featx = spectral_features(x, Fs, feat_name, params)

    Inputs:
        x          - epoch of EEG data (size 1 x N)
        Fs         - sampling frequency (in Hz)
        feat_name  - feature type, defaults to 'spectral_power';
                     see full list of 'spectral_' features in params['FEATURE_SET_ALL']
        params  - parameters (as dictionary);
                     see NEURAL_parameters() for examples

    Outputs:
        featx  - feature at each frequency band

    Example:
        import utils
        import spectral_features
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values

        featx = spectral_features.main_spectral(x, Fs, 'spectral_relative_power')

    """
    if params is None:
        params = NEURAL_parameters.NEURAL_parameters()
        if "spectral" in params:
            params = params["spectral"]
        else:
            raise ValueError("No default parameters found")
    elif len(params) == 0:
        params = NEURAL_parameters.NEURAL_parameters()
        if "spectral" in params:
            params = params["spectral"]
        else:
            raise ValueError("No default parameters found")

    try:
        if feat_name == "spectral_relative_power":
            featx = eval("spectral_power")(x, Fs, feat_name, params)
        else:
            featx = eval(feat_name)(x, Fs, feat_name, params)
    except:
        raise ValueError("Feature function not found: %s" % feat_name)

    return featx



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

import numpy as np
from NEURAL_py_EEG import spectral_features
import warnings
from NEURAL_py_EEG import utils
from scipy import signal
from scipy import stats
from NEURAL_py_EEG import NEURAL_parameters
import matplotlib as plt


def mat_percentile(x, q):
    """
    Percentile function to mimic MATLAB
    :param x:
    :param q:
    :return:
    """
    q = np.array(q) / 100
    x = np.delete(x, np.where(np.isnan(x))[0])
    n = len(x)
    y = np.sort(x)
    return np.interp(q, np.linspace(1 / (2 * n), (2 * n - 1) / (2 * n), n), y)


def gen_cross_spectrum(x, y, Fs, param_st):
    """

    gen_cross_spectrum: cross-spectrums

    Syntax:
        [pxy,Nfreq,f_scale,fp] = connectivity_features.gen_cross_spectrum(x,y,Fs,param_st)

    Inputs:
        x,y,Fs,param_st -

    Outputs:
        [pxy,Nfreq,f_scale,fp]

    Example:

    """
    L_window = param_st["L_window"]
    window_type = param_st["window_type"]
    overlap = param_st["overlap"]
    # freq_bands = param_st['freq_bands']
    spec_method = param_st["method"]

    # remove nans
    x[np.isnan(x)] = []
    if spec_method.lower() == "bartlett-psd":
        # ---------------------------------------------------------------------
        # Bartlett PSD: same as Welch with 0 % overlap and rectangular window
        # ---------------------------------------------------------------------
        window_type = "rect"
        overlap = 0

        spec_method = "psd"

    # ---------------------------------------------------------------------
    # Welch cross - PSD
    # ---------------------------------------------------------------------
    S_x, Nfreq, f_scale, win_epoch = spectral_features.gen_STFT(
        x, L_window, window_type, overlap, Fs, 1
    )
    S_y, Nfreq, f_scale, win_epoch = spectral_features.gen_STFT(
        y, L_window, window_type, overlap, Fs, 1
    )

    S_xy = S_x * np.conjugate(S_y)  # check this

    if spec_method.lower() == "psd":
        # ---------------------------------------------------------------
        # Mean for Welch PSD
        # ---------------------------------------------------------------
        pxy = np.nanmean(S_xy, axis=0)  # Check this
    elif spec_method.lower() == "robust-psd":
        # ---------------------------------------------------------------
        # Mean for Welch PSD
        # ---------------------------------------------------------------
        pxy = np.nanmedian(S_xy, axis=0)  # Check this

    else:
        raise ValueError(
            "Unknown cross-spectral method %s - check spelling" % spec_method
        )

    # normalise (so similar to Welch's PSD)
    E_win = np.sum(np.abs(win_epoch) ** 2) / Nfreq
    pxy = pxy / (Nfreq * E_win * Fs)

    N = len(pxy)
    f_scale = Nfreq / Fs

    fp = np.array(range(N)) / f_scale

    return pxy, Nfreq, f_scale, fp


def gen_coherence(x, y, Fs, param_st):
    """
    ---------------------------------------------------------------------
     generate coherence (magnitude only) between x and y
    ---------------------------------------------------------------------
    :param x:
    :param y:
    :param Fs:
    :param param_st:
    :return:
    """
    pxx = spectral_features.gen_spectrum(x, Fs, param_st)[0]
    pyy = spectral_features.gen_spectrum(y, Fs, param_st)[0]
    pxy, dum, f_scale, fp = gen_cross_spectrum(x, y, Fs, param_st)

    c = (np.abs(pxy) ** 2) / (pxx * pyy)
    return c, pxx, pyy, f_scale, fp


def rand_phase(x, N_iter=1):
    """

    Syntax: y = connectivity_features.rand_phase(x, N_iter)

    Inputs:
        x      - input signal, length-N
        N_iter - number of random signals required

    Outputs:
        y - output signal, same amplitude spectrum but with random phase
            (size N_iter x N)

    Example:
        import numpy as np
        import connectivity_features
        import matplotlib.pyplot as plt

        N = 1000
        N_iter = 500
        x = np.random.randn(N)
        y = connectivity_features.rand_phase(x, N_iter)

        plt.figure(1)
        plt.subplot(2, 1, 1)
        plt.plot(x)
        plt.subplot(2, 1, 2)
        plt.plot(y[1, :])
        plt.show()


    rand_phase: randomize phase (of FFT) for real-valued signal

    :param x: Input signal of length N
    :param N_iter: Number of random signals required
    :return: y - output signal, same amplitude spectrum but with random phase (size N_iter x N
    """
    if len(x) == 1:
        x = x.transpose()

    N = len(x)  # check
    Nh = np.floor(N / 2).astype(int)

    X = np.fft.fft(x)

    # ---------------------------------------------------------------------
    # 1. generate random phase
    # ---------------------------------------------------------------------
    # np.random.seed(ttt)
    if N % 2 == 1:
        rphase = np.array(-np.pi + 2 * np.pi * np.random.uniform(0, 1, [N_iter, Nh]))
        rphase = np.hstack([np.zeros([N_iter, 1]), rphase[:, :], -rphase[:, ::-1]])
    else:
        rphase = np.array(
            -np.pi + 2 * np.pi * np.random.uniform(0, 1, [N_iter, Nh - 1])
        )
        # rand is to be flipped.
        # rphase = np.array(-np.pi + 2 * np.pi * np.random.uniform(0, 1, [Nh-1, N_iter]))  # Note: to replicate MATLAB
        # rphase = np.transpose(rphase)
        rphase = np.hstack(
            [np.zeros([N_iter, 1]), rphase, np.zeros([N_iter, 1]), -rphase[:, ::-1]]
        )

    # ---------------------------------------------------------------------
    # apply to spectrum and FFT back to time domain:
    # ---------------------------------------------------------------------

    if N_iter > 1:
        Y = np.empty([N_iter, len(X)]).astype(complex)
        for i in range(N_iter):
            for j in range(len(X)):
                Y[i, j] = np.exp(complex(0, rphase[i, j])) * np.abs(X[j])

    else:
        Y = np.empty(len(X)).astype(complex)
        for j in range(len(X)):
            Y[j] = np.abs(X[j]) * np.exp(complex(0, rphase[j]))

    y = np.real(np.fft.ifft(Y, axis=1))

    return y


def connectivity_BSI(
    x,
    Fs,
    params_st,
    N_channels,
    ileft,
    iright,
    ipairs,
    N_freq_bands,
    freq_bands,
    feat_name,
):
    """
    ---------------------------------------------------------------------
     brain sysmetry index (revised version, Van Putten, 2007)

     van Putten, MJAM (2007). The revised brain symmetry index. Clinical
     Neurophysiology, 118(11), 2362â€“2367. http://doi.org/10.1016/j.clinph.2007.07.019
    ---------------------------------------------------------------------
    :param feat_name:
    :param freq_bands:
    :param N_freq_bands:
    :param ipairs:
    :param iright:
    :param ileft:
    :param N_channels:
    :param x: input epoch
    :param Fs: sampling frequency
    :param params_st: parameters
    :return:
    """

    # a) PSD estimate (Welch's periodogram)
    x_epoch = np.zeros([len(x)])
    X = np.zeros([N_channels, len(x_epoch)])
    for k in range(N_channels):
        x_epoch = x[k, :]
        x_epoch = np.delete(x_epoch, np.where(np.isnan(x_epoch))[0])

        tmp, dum, f_scale, dum, fp = spectral_features.gen_spectrum(
            x_epoch, Fs, params_st, 1
        )
        if k == 0:
            X = np.zeros([N_channels, len(tmp)])
        X[k, :] = tmp

    dum, N = X.shape  # Double check this

    if len(ileft) > 1:
        X_left = np.nanmean(
            X[ipairs[0, :].astype(int), :], axis=0
        )  # Check what the 1 does
        X_right = np.nanmean(X[ipairs[1, :].astype(int), :], axis=0)  # Again check
    else:
        X_left = X[ileft, :]
        X_right = X[iright, :]

    featx = np.zeros(N_freq_bands)
    ibandpass = []  # to suppress warning
    for n in range(N_freq_bands):
        if n == 0:
            istart = np.ceil(freq_bands[n, 0] * f_scale).astype(int) - 1
        else:
            istart = ibandpass[-1] - 1

        ibandpass = np.array(
            range(istart, np.floor(freq_bands[n][1] * f_scale).astype(int))
        )
        ibandpass = ibandpass + 1
        ibandpass[ibandpass < 0] = 0
        ibandpass[ibandpass > N - 1] = N - 1

        featx[n] = np.nanmean(
            np.abs(
                (X_left[ibandpass] - X_right[ibandpass])
                / (X_left[ibandpass] + X_right[ibandpass])
            )
        )

    return featx


def connectivity_coh(
    x,
    Fs,
    params_st,
    N_channels,
    ileft,
    iright,
    ipairs,
    N_freq_bands,
    freq_bands,
    feat_name,
):
    """
    ---------------------------------------------------------------------
     coherence (using Welch's PSD)
    ---------------------------------------------------------------------

    :param x:
    :param Fs:
    :param params_st:
    :param N_channels:
    :param ileft:
    :param iright:
    :param ipairs:
    :param N_freq_bands:
    :param freq_bands:
    :return:
    """
    # check to see if using the right PSD estimate
    # (only 'PSD' or 'bartlett-PSD' allowed)
    if params_st["method"] == "periodogram" or params_st["method"] == "robust-PSD":
        print("----------- -WARNING- ------------\n")
        print("Must use averaging PSD estimate (e.g. Welch PSD) for coherence.\n")
        print('To do so, set: feat_params_st.connectivity.method="PSD"\n')
        print("in " "NEURAL_parameters.py" " file.\n")

        if params_st["coherence_zero_level"].lower() == "analytic":
            params_st["method"] = "bartlett-PSD"

            warnings.warn(
                "Forcing PSD method for connectivity analysis to Bartlett PSD",
                DeprecationWarning,
            )

        else:
            warnings.warn(
                "Forcing PSD method for connectivity analysis to Welch PSD \
            (may need to adjust window type, window size, and overlap.)",
                DeprecationWarning,
            )

            params_st["method"] = "PSD"
        print("------------------------------------------------------------------")

    if (
        params_st["coherence_zero_level"].lower() == "analytic"
        and params_st["method"] != "bartlett-PSD"
    ):

        # check if PSD method is Bartlett or not:
        print("----------- -WARNING- ------------\n")
        print("If want to use the analytic zero-level threshold for\n")
        print("coherence (Halliday et al. 1995) then need to use Barlett PSD.\n")
        print("To do so, set: feat_params_st.connectivity.method='bartlett-PSD'\n")
        print("in 'NEURAL_parameters.py' file.\n")
        warnings.warn(
            ["Forcing PSD method for connectivity analysis to Bartlett PSD"],
            DeprecationWarning,
        )
        print("----------------------------------\n")

        params_st["method"] = "bartlett-PSD"

    # PSD and x-PSD
    dum, N_pairs = ipairs.shape
    featx_pairs = np.empty([N_freq_bands, N_pairs])
    featx_pairs.fill(np.nan)

    coh = np.zeros([N_pairs, 1])

    for p in range(N_pairs):
        x1 = x[ipairs[0, p].astype(int), :]
        x2 = x[ipairs[1, p].astype(int), :]
        x1 = np.delete(x1, np.where(np.isnan(x1))[0])
        x2 = np.delete(x2, np.where(np.isnan(x2))[0])

        # 1) estimage the coherence function
        c, pxx, pyy, f_scale, fp = gen_coherence(
            x1, x2, Fs, params_st
        )  # Check size of coh
        if p == 0:
            coh = np.empty([N_pairs, len(c)])
        coh[p, :] = c

        # 2) if estimating a zero-level threshold for the coherence:
        if params_st["coherence_zero_level"] == "surr":
            # ---------------------------------------------------------------------
            # if generating a null-hypothesis distribution from surrogate data:
            # ---------------------------------------------------------------------

            L_surr = params_st["coherence_surr_iter"]

            # Generate surrogate signals
            x1_surr = rand_phase(x1, L_surr)
            x2_surr = rand_phase(x2, L_surr)

            coh_sur = []  # to suppress warning
            for m in range(L_surr):
                c_sur, dum, dum, dum, dum = gen_coherence(
                    x1_surr[m, :], x2_surr[m, :], Fs, params_st
                )
                if m == 0:
                    coh_sur = np.empty([L_surr, len(c_sur)])
                coh_sur[m, :] = c_sur

            # estimating frequency-dependent threshold at p < Î± level of significance
            coh_threshold = np.empty([1, len(c_sur)])
            for c in range(len(c_sur)):
                coh_threshold[:, c] = mat_percentile(
                    coh_sur[:, c], 100 * (1 - params_st["coherence_zero_alpha"])
                )

        elif params_st["coherence_zero_level"] == "analytic":
            # ---------------------------------------------------------------------
            # or if using an analytic method
            # ---------------------------------------------------------------------
            # number of segments(no overlap with Bartlett PSD)
            L = np.floor(len(x1) / (Fs * params_st["L_window"]))

            coh_threshold = 1 - (params_st["coherence_zero_alpha"]) ** (1 / (L - 1))

        elif params_st["coherence_zero_level"].lower == "":
            coh_threshold = np.array([])
        else:
            raise ValueError(
                "unknown option for 'coherence_zero_level'; see NEURAL_parameters.py for details"
            )

        # 3) threshold

        if coh_threshold.size != 0:
            coh[p, coh[p, :] < coh_threshold] = 0  # Check this

        # compute the coherence (either mean, max, or max. frequency) for each frequency band
        dum, N = coh.shape
        ibandpass = []  # to suppress warning
        for n in range(N_freq_bands):
            if n == 0:
                istart = np.ceil(freq_bands[n, 0] * f_scale).astype(int) - 1
            else:
                istart = ibandpass[-1] - 1

            ibandpass = np.array(
                range(istart, np.floor(freq_bands[n][1] * f_scale).astype(int))
            )  # May need +1
            ibandpass = ibandpass + 1
            ibandpass[ibandpass < 0] = 0
            ibandpass[ibandpass > N - 1] = N - 1

            if feat_name == "connectivity_coh_mean":
                featx_pairs[n, p] = np.nanmean(coh[p, ibandpass])
            elif feat_name == "connectivity_coh_max":
                featx_pairs[n, p] = np.max(coh[p, ibandpass])
            elif feat_name == "connectivity_coh_freqmax":
                imax = np.argmax(coh[p, ibandpass])
                featx_pairs[n, p] = fp[ibandpass[imax]]

    featx = np.nanmedian(featx_pairs, 1)  # compare this
    return featx


def connectivity_corr(
    x,
    Fs,
    params_st,
    N_channels,
    ileft,
    iright,
    ipairs,
    N_freq_bands,
    freq_bands,
    feat_name,
):
    """
    ---------------------------------------------------------------------
     cross-correlation (Pearson)
    ---------------------------------------------------------------------
    :param x:
    :param Fs:
    :param params_st:
    :param N_channels:
    :param ileft:
    :param iright:
    :param ipairs:
    :param N_freq_bands:
    :param freq_bands:
    :param feat_name:
    :return:
    """
    x_orig = x.copy()
    dum, N_pairs = ipairs.shape
    featx = np.empty(N_freq_bands)
    for n in range(N_freq_bands):
        inans = dict()
        x_filt = np.empty([N_channels, x.shape[1]])
        for p in range(N_channels):
            x_filt[p, :], inans[p] = utils.filter_butterworth_withnans(
                x_orig[p, :],
                Fs,
                freq_bands[n][1],
                freq_bands[n][0],
                5,
                params_st["FILTER_REPLACE_ARTEFACTS"],
            )
            x_orig = x.copy()

        cc_pairs = np.empty(N_pairs)
        cc_pairs.fill(np.nan)

        for p in range(N_pairs):
            all_inans = np.unique(
                np.hstack(
                    [inans[ipairs[0, p].astype(int)], inans[ipairs[1, p].astype(int)]]
                )
            )

            x1 = x_filt[ipairs[0, p].astype(int), :]
            x2 = x_filt[ipairs[1, p].astype(int), :]
            if all_inans.size != 0:
                x1 = np.delete(x1, all_inans)
                x2 = np.delete(x2, all_inans)

            env1 = np.abs(signal.hilbert(x1)) ** 2
            env2 = np.abs(signal.hilbert(x2)) ** 2

            cc_pairs[p] = stats.pearsonr(env1, env2)[0]

        featx[n] = np.nanmedian(cc_pairs)

    return featx


def channel_hemispheres(channels_all):
    """
    Syntax: ipairs = channel_hemisphere_pairs(channels_all)

    Inputs:
        channels_all - list of bipolar channel names
                         e.g. ['C3-O1','C4-O2', 'F3-C3', 'F4-C4']

    Outputs:
        ileft  - indices of the channels on left hemispheres
        iright - indices of the channels on right hemispheres

    Example:
        import utils
        import connectivity_features
        import numpy as np
        
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        channel_labels = np.array(data_st['ch_labels'])
        
        ileft, iright = connectivity_features.channel_hemispheres(channel_labels)

    """
    N_channels = len(channels_all)
    ileft = np.array([])
    iright = np.array([])
    for n in range(N_channels):
        cname = channels_all[n]
        M = len(cname)
        il = np.array([])
        ir = np.array([])
        for p in range(M):
            if str.isdigit(cname[p]):
                num = int(cname[p])
                if num % 2 == 1:
                    il = np.append(il, n)
                else:
                    ir = np.append(ir, n)

        if il.size != 0 and ir.size != 0:
            raise ValueError("both odd and even in channel: %s" % cname)

        if il.size != 0:
            ileft = np.append(ileft, n)
        elif ir.size != 0:
            iright = np.append(iright, n)
        else:
            warnings.warn("left or right channel:  %s ?" % cname, DeprecationWarning)

    return ileft.astype(int), iright.astype(int)


# noinspection SpellCheckingInspection
def channel_hemisphere_pairs(channel_labels):
    """
    Inputs:
        channel_labels - list of bipolar channel names
                         e.g. ['C3-O1','C4-O2', 'F3-C3', 'F4-C4']

    Outputs:
        ipairs - indices of the channel-pairs across hemispheres
                 e.g. 'C3-O1' is paired with 'C4-O2', 'F3-C3' paired with 'F4-C4'

    Example:
        import utils
        import connectivity_features
        import numpy as np
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        channel_labels = np.array(data_st['ch_labels'])

        ipairs = connectivity_features.channel_hemisphere_pairs(channel_labels)
        channel_labels[ipairs]

    channel_hemisphere_pairs: pair left/right channels

    Syntax: ipairs=channel_hemisphere_pairs(channel_labels)
    """
    DBverbose = 0
    # N = len(channel_labels)
    ileft, iright = channel_hemispheres(channel_labels)
    N_left = len(ileft)

    ipairs = np.empty([2, N_left])
    ipairs.fill(np.nan)
    [channel_labels[xx].upper() for xx in range(len(channel_labels))]

    for n in range(N_left):
        ipairs[0, n] = ileft[n]
        ch_left = channel_labels[ileft[n]].upper()

        # change numbers from odd to even
        ch_left_match = str.replace(
            str.replace(
                str.replace(str.replace(ch_left, "1", "2"), "3", "4"), "5", "6"
            ),
            "7",
            "8",
        )  # check this

        imatch = np.array([])
        for i in range(len(iright)):
            if channel_labels[iright[i]] == ch_left_match:
                imatch = np.array(i)
                break
        # and check for reversed order
        sep = str.find(ch_left_match, "-")
        ch1 = ch_left_match[:sep]
        ch2 = ch_left_match[sep + 1 :]
        ch_left_match_rv = ch2 + "-" + ch1

        imatch_rv = np.array([])
        for i in range(len(iright)):
            if channel_labels[iright[i]] == ch_left_match_rv:
                imatch_rv = np.array(i)
                break

        if imatch.size != 0:
            ipairs[1, n] = iright[imatch]
        elif imatch_rv.size != 0:
            ipairs[1, n] = iright[imatch_rv]
        else:
            ipairs[0:1, n] = np.nan
            if DBverbose:
                print("no matching pair for channel: %s\n" % ch_left)

        # if left / right side share common electrode (e.g.Cz),
        # then should ignore
        if ch1 in ch_left or ch2 in ch_left:
            ipairs[0:1, n] = np.nan

    irem = []
    for n in range(N_left):
        if np.isnan(ipairs[:, n]).any():
            irem.append(n)

    if len(irem) != 0:
        ipairs = np.delete(ipairs, irem, axis=1)
        if DBverbose:
            print(channel_labels[ipairs])

    return ipairs.astype(int)


def main_connectivity(x, Fs, feat_name, ch_labels, params=None):
    """
    Syntax: featx = connectivity_features(x, Fs, feat_name, params)

    Inputs:
        x          - epoch of EEG data (size 1 x N)
        Fs         - sampling frequency (in Hz)
        feat_name  - feature type, defaults to 'connectivity_BSI';
                     see full list of 'connectivity_' features in params['FEATURE_SET_ALL']
        params  - parameters (as dictionary)
                     see NEURAL_parameters() for examples

    Outputs:
        featx  - feature at each frequency band

    Example:
        import utils
        import connectivity_features
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'].to_numpy().transpose()
        channel_labels = data_st['ch_labels']

        featx = connectivity_features.main_connectivity(x, Fs, 'connectivity_corr', channel_labels)


    [1] van Putten, MJAM (2007). The revised brain symmetry index. Clinical Neurophysiology,
        118(11), 2362â€“2367.
    [2] Prichard D, Theiler J (1994). Generating surrogate data for time series with several
        simultaneously measured variables. Physical Review Letters, 1994;73(7):951â€“954.
    [3] Faes L, Pinna GD, Porta A, Maestri R, Nollo G (2004). Surrogate data analysis for
        assessing the significance of the coherence function. IEEE Transactions on
        Biomedical Engineering, 51(7):1156â€“1166.
    [4] Halliday, DM, Rosenberg, JR, Amjad, AM, Breeze, P, Conway, BA, &
        Farmer, SF. (1995). A framework for the analysis of mixed time series/point
        process data--theory and application to the study of physiological tremor, single
        motor unit discharges and electromyograms. Progress in Biophysics and Molecular
        Biology, 64(2â€“3), 237â€“278.

    """
    if params is None:
        params = NEURAL_parameters.NEURAL_parameters()
        if "connectivity" in params:
            params = params["connectivity"]
        else:
            raise ValueError("No default parameters found")
    elif len(params) == 0:
        params = NEURAL_parameters.NEURAL_parameters()
        if "connectivity" in params:
            params = params["connectivity"]
        else:
            raise ValueError("No default parameters found")
    if "connectivity" in params:
        params = params["connectivity"]

    freq_bands = np.array(params["freq_bands"])

    N_channels, dum = x.shape

    if N_channels < 2:
        warnings.warn("Requires at least 2 channels", DeprecationWarning)
        return np.nan

    N_freq_bands, dum = freq_bands.shape
    if N_freq_bands == 0:
        N_freq_bands = 1

    ileft = []
    iright = []
    ipairs = []
    if N_channels > 2 and len(ch_labels) != 0:
        ileft, iright = channel_hemispheres(ch_labels)
        ipairs = channel_hemisphere_pairs(ch_labels)
    elif N_channels == 2:
        # if no channel labels then guess
        ileft = 0
        iright = 1
        ipairs = np.array([1, 2])

    if ipairs.shape[1] == 0:
        return np.array([np.nan, np.nan, np.nan, np.nan])
    try:
        if (
            feat_name == "connectivity_coh_mean"
            or feat_name == "connectivity_coh_max"
            or feat_name == "connectivity_coh_freqmax"
        ):

            featx = eval("connectivity_coh")(
                x,
                Fs,
                params,
                N_channels,
                ileft,
                iright,
                ipairs,
                N_freq_bands,
                freq_bands,
                feat_name,
            )
        else:
            featx = eval(feat_name)(
                x,
                Fs,
                params,
                N_channels,
                ileft,
                iright,
                ipairs,
                N_freq_bands,
                freq_bands,
                feat_name,
            )
    except:
        raise ValueError("Feature function not found: %s" % feat_name)

    return featx



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

import math
import numpy as np
from scipy import signal
from scipy.interpolate import interp1d
from scipy.interpolate import PchipInterpolator
import random
import warnings
import matplotlib.pyplot as plt
import pandas as pd
import pyedflib
from NEURAL_py_EEG import mfiltfilt


# noinspection SpellCheckingInspection
def get_window(win_length, win_type, Npad=0):
    """
    USE: win = get_window( win_length, win_type )

    INPUT:
          win_length = length of window - samples
          win_type   = type of window: { 'rect' | 'hamm' | 'hann' | 'tuke' | 'bart' }
          Npad       = (optional) zero-pad window to length Npad.

    OUTPUT:
          win = window

    EXAMPLE:
        import utils
        import matplotlib.pyplot as plt

        N = 64
        win = utils.get_window(N, 'hamm', 100)
        plt.plot(win)
        plt.show()

    (int,str) -> list

    Pass in a window length and window type string and return a window

    """

    win_type = win_type.lower()

    if win_type == "hamm" or win_type == "hamming":
        win = np.hamming(win_length)
    elif win_type == "hann" or win_type == "hanning":
        win = np.hanning(win_length)
    elif win_type == "rect" or win_type == "rectangular":
        win = np.ones(win_length)
    elif win_type == "black" or win_type == "blackman":
        win = np.blackman(win_length)
    elif win_type == "blackmanharris":
        win = signal.blackmanharris(win_length)
    elif win_type == "kaiser":
        win = np.kaiser(win_length, 5)
    elif win_type == "tuke" or win_type == "tukey":
        win = signal.tukey(win_length)
    elif win_type == "bart" or win_type == "bartlett":
        win = signal.bartlett(win_length)
    else:
        raise ValueError("Name of window function incorrect. Name: %s" % win_type)

    win = np.roll(win, math.ceil(len(win) / 2))

    if Npad != 0:
        win = pad_win(win, Npad)

    return win


def pad_win(w, Npad):
    """
    --------------------------------------------------------------------------------

     Pad window to Npad.

     Presume that positive window indices are first.

     When N is even use method described in [1]

       References:
         [1] S. Lawrence Marple, Jr., Computing the discrete-time analytic
         signal via FFT, IEEE Transactions on Signal Processing, Vol. 47,
         No. 9, September 1999, pp.2600--2603.

    --------------------------------------------------------------------------------
    """
    w_pad = np.zeros(Npad)
    N = len(w)
    Nh = np.floor(N / 2).astype(int)
    if Npad < N:
        raise ValueError("Npad is less than N")

    # Trivial case
    if N == Npad:
        return w

    # For N odd:
    if N % 2 == 1:
        n = list(range(Nh + 1))
        w_pad[n] = w[n]
        n = np.array(range(Nh))
        w_pad[Npad - n - 1] = w[N - n - 1]

        # For N even:
        # split the Nyquist frequency in two and distribute over positive
        # and negative indices.
    else:
        n = list(range(Nh))
        w_pad[n] = w[n]
        w_pad[Nh] = w[Nh] / 2

        n = np.array(range(Nh - 1))
        w_pad[Npad - n - 1] = w[N - n - 1]
        w_pad[Npad - Nh] = w[Nh] / 2
    return w_pad


# noinspection SpellCheckingInspection
def gen_epoch_window(L_overlap, L_epoch, win_type, Fs, GEN_PSD=None):
    """
    Syntax: L_hop, L_epoch, win_epoch = gen_epoch_window(L_overlap, L_epoch, win_type, Fs)

    Inputs:
        L_overlap - precentage overlap
        L_epoch   - epoch size (in seconds)
        win_type  - window type, e.g. 'hamm' for Hamming window
        Fs        - sampling frequency (in Hz)

    Outputs:
        L_hop     - hop size (in samples)
        L_epoch   - epoch size (in samples)
        win_epoch - window, of length L_epoch

    Example:
        import utils
        import matplotlib.pyplot as plt
        import numpy as np

        overlap = 50
        win_length = 2
        Fs = 64

        L_hop, L_epoch, win_epoch = utils.gen_epoch_window(overlap, win_length, 'hamm', Fs)

        print('hop length=%d; epoch length=%d' % (L_hop, L_epoch))
        plt.figure(1)
        ttime = np.array(range(len(win_epoch))) / Fs
        plt.plot(ttime, win_epoch)
        plt.show()

    (int, int, str, int) -> int, int, list

    L_hop, L_epoch, win_epoch = gen_epoch_window(win_overlap, win_length, win_type, fs)
    window

    """
    if GEN_PSD is None:
        GEN_PSD = 0

    L_hop = (100 - L_overlap) / 100
    L_epoch = math.floor(L_epoch * Fs)

    if GEN_PSD:
        # ---------------------------------------------------------------------
        # if PSD
        # ---------------------------------------------------------------------
        L_hop = np.ceil((L_epoch - 1) * L_hop)

        win_epoch = get_window(L_epoch, win_type)
        win_epoch = np.roll(win_epoch, math.floor(len(win_epoch) / 2))
    else:
        """ ---------------------------------------------------------------------
        otherwise, if using an overlap-and-add method, then for window w[n]
        âˆ‘â‚˜ w[n - mR] = 1 over all n (R = L_hop )
        
        Smith, J.O. "Overlap-Add (OLA) STFT Processing", in 
        Spectral Audio Signal Processing,
        http://ccrma.stanford.edu/~jos/sasp/Hamming_Window.html, online book, 
        2011 edition, accessed Nov. 2016.
        
        
        there are some restrictions on this:
        e.g. for Hamming window, L_hop = (L_epoch-1)/2, (L_epoch-1)/4, ... 
        ---------------------------------------------------------------------
        """
        if win_type == "hamm":
            L_hop = (L_epoch - 1) * L_hop
        elif win_type == "hann":
            L_hop = (L_epoch + 1) * L_hop
        else:
            L_hop = L_epoch * L_hop

        L_hop = math.ceil(L_hop)
        win_epoch = get_window(L_epoch, win_type)
        win_epoch = np.roll(win_epoch, math.floor(len(win_epoch) / 2))

        if win_type == "hamm" and L_epoch // 2 == 1:
            win_epoch[0] = win_epoch[0] / 2
            win_epoch[-1] = win_epoch[-1] / 2

    return L_hop, L_epoch, win_epoch


def gen_test_EEGdata(dur, Fs, include_bipolar=0, discont_activity=0):
    """
    Syntax: data_st = gen_test_EEGdata(dur, Fs, include_bipolar, discont_activity)

    Inputs:
        dur:              duration of EEG-like data in seconds (default 300 seconds)
        Fs:               sampling frequency
        include_bipolar:  include the bipolar montage aswell as the referential
        discont_activity: discontinuous-like activity of preterm EEG (bursts and inter-bursts)

    Outputs:
        packaged_data:    dictionary including EEG data (referential montage), sampling
                          frequency, and channel labels. Data are pandas DataFrame

    Example:
        import utils
        import matplotlib.pyplot as plt
        import numpy as np
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, discont_activity=1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values

        plt.figure(1)
        ttime = np.array(range(len(x)))/Fs
        plt.plot(ttime, x)
        plt.xlabel('time (seconds)')
        plt.ylabel('$\mu$V')
        plt.show()

    """
    N = int(np.floor(dur * Fs))

    """---------------------------------------------------------------------
    generate white Gaussian noise and filter with moving average filter
    (referential montage has common signal)
    ---------------------------------------------------------------------"""

    L_ma = Fs / 4
    L_ma_noise = np.ceil(Fs / 16)
    N_channels = 9

    eeg_common = 5 * signal.lfilter(
        [1 / L_ma for a in range(int(L_ma))], 1, np.random.randn(int(N))
    )

    eeg_data_ref = (
        signal.lfilter(
            [1 / L_ma for a in range(int(L_ma))], 1, np.random.randn(N_channels, N)
        )
        + (
            signal.lfilter(
                [1 / L_ma_noise for a in range(int(L_ma_noise))],
                1,
                np.random.randn(N_channels, N),
            )
        )
        / 100
        + eeg_common
    )

    if discont_activity:
        ibursts = np.array(
            [random.randint(0, N - 1) for a in range(int(dur / 100) * Fs)]
        )
        amps = np.abs(1 + np.random.uniform(0, 1, [1, int(dur / 100) * Fs]))

        bursts = np.zeros([1, N])
        bursts[0][ibursts] = 1 * amps

        L_win = Fs * 2
        bursts_smooth = 100 * signal.lfilter(np.hamming(L_win) / L_win, 1, bursts)

        eeg_data_ref = 10 * (eeg_data_ref + (eeg_data_ref * (20 * bursts_smooth)))

    else:
        eeg_data_ref = eeg_data_ref * 50

    ch_labels_bi = [
        ["F4", "C4"],
        ["F3", "C3"],
        ["C4", "T4"],
        ["C3", "T3"],
        ["C4", "Cz"],
        ["Cz", "C3"],
        ["C4", "O2"],
        ["C3", "O1"],
    ]

    ch_labels_ref = ["C3", "C4", "Cz", "F3", "F4", "O1", "O2", "T3", "T4"]

    eeg_data_ref = pd.DataFrame(data=np.transpose(eeg_data_ref), columns=ch_labels_ref)
    packaged_data = {
        "eeg_data_ref": eeg_data_ref,
        "Fs": Fs,
        "ch_labels_ref": list(eeg_data_ref.columns),
    }

    if include_bipolar:
        packaged_data["ch_labels_bi"] = ch_labels_bi
        eeg_data, dum = set_bi_montage(eeg_data_ref, Fs, ch_labels_bi)
        packaged_data["eeg_data"] = eeg_data
        packaged_data["ch_labels"] = list(eeg_data.columns)

    return packaged_data


def set_bi_montage(sigs, Fs, bi_mont=None):
    """
    set_bbiploar_montage: Convert monopolar (referential) to bi-polar montgage

    Syntax: bi_sigs, bi_labels = set_bi_montage(sigs,channel_names)

    Inputs:
        sigs          - EEG data in referential montage - pd dataframe
        channel_names - cell of referential channel names
                         e.g. ['C3','C4','F3','F4']
        bi_mont - Bipolar montage

    Outputs:
        bi_sigs   - EEG data in referential montage
        bi_labels - cell of bipolar channel names
                    e.g. ['C3-O1','C4-O2', 'F3-C3', 'F4-C4']

    Example:
        import utils
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)

        x_ref = data_st['eeg_data_ref']
        ch_labels_ref = data_st['ch_labels_ref']
        BI_MONT=[['F4','C4'], ['F3','C3'], ['C4','T4'], ['C3','T3'], ['C4','Cz'], ['Cz','C3'], ['C4','O2'], ['C3','O1']]

        x_bi, ch_labels_bi = utils.set_bi_montage(x_ref, ch_labels_ref, BI_MONT)
    """
    if bi_mont is None:
        bi_mont = [
            ["C3", "O1"],
            ["C4", "O2"],
            ["Cz", "C3"],
            ["C4", "Cz"],
            ["C3", "T3"],
            ["C4", "T4"],
            ["F3", "C3"],
            ["F4", "C4"],
        ]

    channel_names = list(sigs.columns)
    montage_data = pd.DataFrame()
    montage_channels = []
    loop_index = 0

    for pair in range(len(bi_mont)):
        if bi_mont[pair][0] in channel_names and bi_mont[pair][1] in channel_names:
            tmp = bi_mont[pair][0] + "_" + bi_mont[pair][1]
            montage_channels.append(tmp)
            tmp_data = sigs[bi_mont[pair][0]] - sigs[bi_mont[pair][1]]
            montage_data.insert(
                loop_index, bi_mont[pair][0] + "_" + bi_mont[pair][1], tmp_data
            )
            loop_index += 1

    return montage_data, montage_channels


def filter_zerophase(
    x, Fs=1, LP_fc=-1, HP_fc=-1, L_filt=0, win_type="hamming", DBplot=0
):
    """
    filter_zerophase: Simple implementation of zero-phase FIR filter using 'filtfilt'

    Syntax: x_filt = filter_zerophase(x, Fs, LP_fc, HP_fc, L_filt)

    Inputs:
        x      - input signal - 1 channel
        Fs     - sample frequency (Hz)
        LP_fc  - lowpass cut off (Hz)
        HP_fc  - highpass cut off (Hz)
        L_filt - length of filter (in samples) - filter order

    Outputs:
        y - filtered signal

    Example:
        import utils
        import matplotlib.pyplot as plt
        import numpy as np

        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values
        LP_fc = 20
        HP_fc = 1

        y = utils.filter_zerophase(x, Fs, LP_fc, HP_fc, 501)

        plt.figure(1)
        ttime = np.array(range(len(x))) / Fs
        plt.plot(ttime, x, 'r')
        plt.plot(ttime, y, 'b')
        plt.show()

    """
    # replace NaNs with zeros
    inans = np.array(np.where(np.isnan(x)))[0]
    if inans.size != 0:
        x[inans] = 0

    N = len(x)

    if L_filt == 0 or L_filt > (N / 4):
        L_filt = make_odd(np.round(N / 4 - 2))

    # ---------------------------------------------------------------------
    # either bandpass, low - pass, or high - pass
    # ---------------------------------------------------------------------

    if LP_fc != -1 and HP_fc != -1:
        if LP_fc > HP_fc:
            # band-pass
            filt_passband = [HP_fc / (Fs / 2), LP_fc / (Fs / 2)]
            b = signal.firwin(L_filt, filt_passband, window=win_type)
        else:
            # or band-stop
            filt_passband = [LP_fc / (Fs / 2), HP_fc / (Fs / 2)]
            b = signal.firwin(L_filt, filt_passband, window=win_type)

    elif LP_fc != -1:
        filt_passband = LP_fc / (Fs / 2)
        b = signal.firwin(L_filt, filt_passband, window=win_type)

    elif HP_fc != -1:
        L_filt = L_filt + 1
        filt_passband = HP_fc / (Fs / 2)
        b = signal.firwin(L_filt - 1, filt_passband, window=win_type)

    else:
        raise ValueError("Need to specify cutoff frequency")

    # Do the filtering

    x_filt = mfiltfilt.mat_filtfilt(b, np.array([1]), x)
    x_filt = x_filt[0:N]

    # Add the nans back in
    if inans.size != 0:
        x_filt[inans] = np.nan

    return x_filt


def make_odd(x):
    x = np.floor(x)
    if x % 2 == 0:
        x = x - 1
    return int(x)


def gen_window(win_type, L):
    # ---------------------------------------------------------------------
    # window
    # ---------------------------------------------------------------------
    win_type = win_type.lower()

    if win_type == "hamm" or win_type == "hamming":
        w = np.hamming(L)
    elif win_type == "hann" or win_type == "hanning":
        w = np.hanning(L)
    elif win_type == "rect" or win_type == "rectangular":
        w = np.ones(L)
    elif win_type == "black" or win_type == "blackman":
        w = np.blackman(L)
    elif win_type == "blackmanharris":
        w = signal.blackmanharris(L)
    elif win_type == "kaiser":
        w = np.kaiser(L, 5)
    else:
        raise ValueError("Name of window function incorrect. Name: %s" % win_type)

    return w


# noinspection SpellCheckingInspection
def filter_butterworth_withnans(
    x, Fs, F3db_lowpass, F3db_highpass, order, FILTER_REPLACE_ARTEFACTS="linear_interp"
):
    """
    Syntax: y, inans = filter_butterworth_withnans(x, Fs, F3db_lowpass, F3db_highpass, order, FILTER_REPLACE_ARTEFACTS)

    Inputs:
        x             - input signal
        Fs            - sample frequency (Hz)
        F3db_lowpass  - 3 dB lowpass cut off (Hz)
        F3db_highpass - 3 dB highpass cut off (Hz)
        order         - filter order
        FILTER_REPLACE_ARTEFACTS - what to do with NaNs?
            either replace with 0s ('zeros'), linear interpolation
            ('linear_interp', default), or cubic spline interpolation ('cubic_interp')

    Outputs:
        y - filtered signal
        inans - locations of nans

    Example:
        import utils
        import matplotlib.pyplot as plt
        import numpy as np

        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values
        F3db_lowpass = 30
        F3db_highpass = 0.5

        y, inans = utils.filter_butterworth_withnans(x, Fs, F3db_lowpass, F3db_highpass, 5)

        plt.figure(1)
        ttime = np.array(range(len(x))) / Fs
        plt.plot(ttime, x)
        plt.plot(ttime, y)
        plt.show()

    """
    inans = np.array([])

    if not F3db_highpass:
        b, a = signal.butter(order, F3db_lowpass / (Fs / 2), "lowpass")
    elif not F3db_lowpass:
        b, a = signal.butter(order, F3db_highpass / (Fs / 2), "highpass")
    else:
        if isinstance(order, list):
            order_low = order[0]
            order_high = order[1]
        else:
            order_low = order
            order_high = order

        y, isnans_low = filter_butterworth_withnans(
            x, Fs, F3db_lowpass, [], order_low, FILTER_REPLACE_ARTEFACTS
        )
        y, isnans_high = filter_butterworth_withnans(
            y, Fs, [], F3db_highpass, order_high, FILTER_REPLACE_ARTEFACTS
        )
        inans = np.unique(np.concatenate([isnans_low, isnans_high]))
        return y, inans

    # remove NaNs and replace with ?
    inans = np.array(np.argwhere(np.isnan(x)))
    if inans.size != 0:
        if FILTER_REPLACE_ARTEFACTS == "zeros":
            x[inans] = 0
        elif FILTER_REPLACE_ARTEFACTS == "linear_interp":
            x = replace_start_ends_NaNs_with_zeros(x)
            x, dum = naninterp(x, "linear")
        elif (
            FILTER_REPLACE_ARTEFACTS == "cubic_interp"
            or FILTER_REPLACE_ARTEFACTS == "nans"
        ):
            x = replace_start_ends_NaNs_with_zeros(x.copy())
            x, dum = naninterp(x.copy(), "pchip")

    y = mfiltfilt.mat_filtfilt(np.array(b), np.array(a), x)

    # special case if nans
    if FILTER_REPLACE_ARTEFACTS.lower() == "nans":
        if inans.size != 0:
            y[inans] = np.nan

    return y, inans


def replace_start_ends_NaNs_with_zeros(x):
    """
    ---------------------------------------------------------------------
     replace leading or trailing NaNs with zeros (needed for naninterp)
    ---------------------------------------------------------------------
    """
    N = len(x)
    istart = np.argwhere(~np.isnan(x))[0][0]
    iend = np.argwhere(~np.isnan(x))[-1][0]

    if istart.size > 0 and istart > 0:
        x[0 : istart + 1] = 0
    if iend.size > 0 and iend < N - 1:
        x[iend + 1 : N + 1] = 0

    return x


def naninterp(X, method="linear"):
    """
    ---------------------------------------------------------------------
     fill 'gaps' in data (marked by NaN) by interpolating
    ---------------------------------------------------------------------

    """
    inan = np.argwhere(np.isnan(X))
    if inan.size == 0:
        return X, inan
    elif len(inan) == 1:
        if inan > 0:
            X[inan] = X[inan - 1]
        else:
            X[inan] = X[inan + 1]
    else:
        try:
            if method != "pchip":
                set_interp = interp1d(
                    np.transpose(np.argwhere(~np.isnan(X)))[0],
                    np.transpose(X[np.argwhere(~np.isnan(X))])[0],
                    kind=method,
                )
                X[inan] = set_interp(inan)
            else:
                set_interp = PchipInterpolator(
                    np.transpose(np.argwhere(~np.isnan(X)))[0],
                    np.transpose(X[np.argwhere(~np.isnan(X))])[0],
                    extrapolate=False,
                )
                X[inan] = set_interp(inan)

        except:
            raise ValueError("linear interpolation with NaNs")
    return X, inan


def len_cont_zeros(x, const=0):
    """
    len_cont_zeros: find length of continuous segments of zeros from binary mask x. Can contain NaNs.

    Syntax: [lens,istart,iend]=len_cont_zeros(x,conts)

    Inputs:
        x     - binary [0,1] vector - 1 row
        const - which to look for, either 0 (default) or 1

    Outputs:
        lens   - array of lengths of segments
        istart - indices: start of segments
        iend   - indices: end of segments

    Example:
        import utils
        import numpy as np

        u=np.zeros([256])
        u[49:130] = 1
        u[204:240] = 1
        u = u.astype(int)
        lens, istart, iend = utils.len_cont_zeros(u,1)

        print('%d segments of 1''s of length: \n' %len(lens) )
    """
    nan_loc = np.where(np.isnan(x))[0]
    DBplot = 0
    if np.array_equal(x, x.astype(bool)) or const not in [0, 1]:
        warnings.warn("Must be a binary signal", DeprecationWarning)

    if const == 1:
        y = np.abs(x - 1)
    else:
        y = x

    # Find run of zeros
    y = (y == 0).astype(float)
    y[nan_loc] = 0
    iedge = np.diff(np.concatenate(([0], y, [0])))
    istart = np.array(np.where(iedge == 1))[
        0
    ]  # Row zero only - input array should be 1d
    iend = np.subtract(np.where(iedge == -1), 1)[0]
    lens = np.array([iend - istart])[0]

    if DBplot:
        plt.figure(100)
        plt.plot(x)
        plt.plot(istart, np.array(x[istart]), marker="x")
        plt.plot(iend, np.array(x[iend]), marker="o", markerfacecolor="none")

    return np.array(lens), istart, iend


def save_pd_as_file(data_to_save, file_name, Fs=None):
    """
    (DataFrame, str) ->

    Pass in a pandas dataFrame and a file name. Then save as csv file in same directory

    > save_pd_as_file(pd, file_name)

    """
    if Fs is not None:
        file_name = file_name + "_Fs_" + str(Fs) + ".csv"
    else:
        file_name = file_name + ".csv"
    if isinstance(data_to_save, pd.DataFrame):
        data_to_save.to_csv(file_name, index=False)
    else:
        raise ValueError("Data needs to be pandas DataFrame to be saved!")


def save_pd_as_edf(data, file_save_name, original_edf_file=None, Fs=250):
    """

    save_pd_as_edf: Save pandas EEG dataframe as edf - nans are converted to zero.

    Syntax: save_pd_as_edf(data, file_save_name, original_edf_file, Fs)

    Inputs:
        data            - Pandas DataFrame containing the data. Column names correspond to channels
        file_save_name  - Name to save newly created EDF file as
        original_edf_file   - Original EDF files file (if present)
        Fs              - Sampling frequency (Hz)

    Outputs:

    Example:
        import utils
        Fs = 256
        data_st = utils.gen_test_EEGdata(5 * 60, Fs, 1)  # 5 minutes of fake data
        eeg_ref = data_st['eeg_data_ref']

        utils.save_pd_as_edf(eeg_ref, 'demo_data.edf', Fs=Fs)

    Note: Saving an edf using pyEDFlib will slightly change the values as it interpolates the signal from physical
          to digital values using the physical and digital maximum and minimum values (hardcoded below).
          The general shape should stay the same.
    """
    if file_save_name[-4:] != ".edf":
        file_save_name = file_save_name + ".edf"

    if original_edf_file is not None and original_edf_file[-4:] != ".edf":
        original_edf_file = original_edf_file + ".edf"

    channel_info = []
    data_list = []
    samples_in_file = []

    if original_edf_file is not None:
        f = pyedflib.EdfReader(original_edf_file)
        loc = f.getSignalLabels().index("F4")
        channels = list(data.columns)
        for ch in channels:
            ch_dict = dict()

            ch_dict["physical_min"] = f.getPhysicalMinimum(loc)
            ch_dict["physical_max"] = f.getPhysicalMaximum(loc)
            ch_dict["digital_min"] = f.getDigitalMinimum(loc)
            ch_dict["digital_max"] = f.getDigitalMaximum(loc)
            ch_dict["dimension"] = f.getPhysicalDimension(loc)
            ch_dict["sample_rate"] = f.getSampleFrequency(loc)
            ch_dict["label"] = ch
            ch_dict["transducer"] = f.getTransducer(loc)
            ch_dict["prefilter"] = f.getPrefilter(loc)

            samples_in_file.append(data.shape[0])
            x = data[ch].to_numpy()
            x[np.where(np.isnan(x))[0]] = 0
            data_list.append(x)
            channel_info.append(ch_dict)
    else:
        channels = list(data.columns)
        for ch in channels:
            ch_dict = dict()
            ch_dict["physical_min"] = -5482.288
            ch_dict["physical_max"] = 5482.288
            ch_dict["digital_min"] = -32767
            ch_dict["digital_max"] = 32767
            ch_dict["dimension"] = "uV"
            ch_dict["sample_rate"] = Fs
            ch_dict["label"] = ch
            ch_dict["transducer"] = ""
            ch_dict["prefilter"] = ""

            samples_in_file.append(data.shape[0])
            x = data[ch].to_numpy()
            x[np.where(np.isnan(x))[0]] = 0
            data_list.append(x)
            channel_info.append(ch_dict)

    w = pyedflib.EdfWriter(
        file_save_name, len(list(data.columns)), file_type=pyedflib.FILETYPE_EDFPLUS
    )
    w.setSignalHeaders(channel_info)
    w.writeSamples(data_list)
    w.close()



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

from NEURAL_py_EEG import utils
import os
import numpy as np
import matplotlib.pyplot as plt
import sys
from pathlib import Path


def mat_percentile(x, q):
    """
    Percentile function to mimic MATLAB
    :param x:
    :param q:
    :return:
    """
    q = np.array(q) / 100
    x = np.delete(x, np.where(np.isnan(x))[0])
    n = len(x)
    if n == 0:
        return np.nan
    y = np.sort(x)
    return np.interp(q, np.linspace(1 / (2 * n), (2 * n - 1) / (2 * n), n), y)


def estimate_IBI_lengths(anno, percentiles_all, Fs):
    """
    Estimate the max./median IBI length
    """
    min_ibi_interval = Fs / 4
    lens_anno = utils.len_cont_zeros(anno)[0]

    ishort = np.where(lens_anno < min_ibi_interval)[0]

    if ishort.size > 0:
        lens_anno = np.delete(lens_anno, ishort)

    if len(lens_anno != 0):
        pc_anno = mat_percentile(lens_anno, percentiles_all)
    else:
        pc_anno = np.zeros(1)
    return pc_anno / Fs


def main_IBI(x, Fs, feat_name):
    """
    IBI_features: burst/inter-burst interval features (assuming preterm EEG, <32 weeks GA)

    Syntax: featx = main_IBI(x, Fs, feat_name, params_st)

    Inputs:
        x          - epoch of EEG data (size 1 x N)
        Fs         - sampling frequency (in Hz)
        feat_name  - feature type, defaults to 'IBI_length_max';
                     see full list of 'IBI_' features in params['FEATURE_SET_ALL']
        params_st  - parameters (as structure);
                     see NEURAL_parameters() for examples

    Outputs:
        featx  - IBI features

    Example:
        import utils
        import IBI_features
        import matplotlib.pyplot as plt
        sys.path.append('path to ->\\py_burst_detector-master\\')
        from burst_detector import utils as bdutils
        N = 5000
        Fs = 64
        x = bdutils.gen_impulsive_noise(N)

        featx = IBI_features.main_IBI(x,Fs,'IBI_burst_number');

    [1] JM Oâ€™ Toole, GB Boylan, RO Lloyd, RM Goulding, S Vanhatalo, NJ Stevenson
    (2017). Detecting Bursts in the EEG of Very and Extremely Premature Infants using a
    Multi-Feature Approach. under review, 2017.

    """

    DBplot = 0

    bdetect_path_string = (
        "Enter path to ->" + "py_burst_detector-master/py_burst_detector-master/"
    )
    bdetect_path = Path(bdetect_path_string)

    bdetect_path_exist = os.path.isfile(
        Path(bdetect_path_string + "burst_detector/" + "eeg_burst_detector.py")
    )
    if not bdetect_path_exist:
        print("\n** ------------ **\n")
        print("Burst detector not included in path (eeg_burst_detector.py)\n\n")
        print("If installed, ensure eeg_burst_detector.py and associated files \n")
        print("are included in the python path.\n")
        print(
            "To do this, update the bdetect_path with the location of the burst detector.\n"
        )
        print("(for more on search paths see: ")
        print(
            [
                "<a href=https://docs.python.org/3/library/os.path.html>python os path</a>)\n\n"
            ]
        )
        print("If the burst detector is not installed, download from:\n")
        print(
            [
                "<a href=http://otoolej.github.io/code/py_burst_detector/> py burst detector source code </a>\n"
            ]
        )
        print("** ------------ **\n")

        return np.nan

    sys.path.append(str(bdetect_path))
    from burst_detector import eeg_burst_detector

    burst_anno, svm_out = eeg_burst_detector.eeg_bursts(x.copy(), Fs)

    if DBplot:
        plt.figure()
        ax1 = plt.subplot(2, 1, 1)
        plt.plot(x)
        plt.subplot(2, 1, 2, sharex=ax1)
        plt.plot(burst_anno)
        plt.ylim((-0.2, 1.2))

    if feat_name == "IBI_length_max":
        """
        ---------------------------------------------------------------------
        max. (95th percentile) inter-burst interval
        ---------------------------------------------------------------------
        """
        return estimate_IBI_lengths(burst_anno, 95, Fs)

    elif feat_name == "IBI_length_median":
        """
        ---------------------------------------------------------------------
        median inter-burst interval
        ---------------------------------------------------------------------
        """
        return estimate_IBI_lengths(burst_anno, 50, Fs)

    elif feat_name == "IBI_burst_prc":
        """
        ---------------------------------------------------------------------
        percentage of bursts
        ---------------------------------------------------------------------
        """
        return (np.nansum(burst_anno) / burst_anno.size) * 100

    elif feat_name == "IBI_burst_number":
        """
        ---------------------------------------------------------------------
        number of bursts
        ---------------------------------------------------------------------
        """
        return utils.len_cont_zeros(burst_anno, 1)[0].size

    else:
        raise ValueError('\n Unknown feature "%s"; check spelling\n' % feat_name)



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

from NEURAL_py_EEG import utils
import math
import numpy as np
from NEURAL_py_EEG import NEURAL_parameters


def mat_percentile(x, q):
    """
    Percentile function to mimic MATLAB
    :param x:
    :param q:
    :return:
    """
    q = np.array(q) / 100
    x = np.delete(x, np.where(np.isnan(x))[0])
    n = len(x)
    y = np.sort(x)
    return np.interp(q, np.linspace(1 / (2 * n), (2 * n - 1) / (2 * n), n), y)


def gen_rEEG(data, win_overlap, win_length, win_type, fs, APPLY_LOG_LINEAR_SCALE):
    """
    (list) -> list

    generate the rEEG from EEG data that has been passed in

    >gen_rEEG()

    Note important to load and use parameters correctly, win_length is 64 seconds for the other features but the
    win_length for this refers to 2 seconds by default

    """
    L_hop, L_epoch, win_epoch = utils.gen_epoch_window(
        win_overlap, win_length, win_type, fs
    )
    N = len(data)
    N_epochs = math.floor((N - (L_epoch - L_hop)) / L_hop)
    if N_epochs < 1:
        N_epochs = 1

    nw = np.array(list(range(0, L_epoch)))
    reeg = np.empty([N_epochs])
    for k in range(N_epochs):
        nf = (nw + k * L_hop) % N
        x_epoch = np.multiply(data[nf], win_epoch)

        if all(np.isnan(x_epoch)):
            reeg[k] = np.nan
        else:
            reeg[k] = np.nanmax(x_epoch) - np.nanmin(x_epoch)
    # no need to resample (as per [1])

    if APPLY_LOG_LINEAR_SCALE:
        ihigh = [idx for idx, val in enumerate(reeg) if val > 50]
        if not ihigh:
            reeg = [
                (50 * math.log(aa) / math.log(50)) if i in ihigh else aa
                for i, aa in enumerate(reeg)
            ]

    return reeg


def rEEG_mean(reeg):
    """
    (list) -> float

    pass in the reeg and return the mean value

    >rEEG_mean([2,3,4])
    3
    
    """
    return np.nanmean(reeg)


def rEEG_median(reeg):
    """
    (list) -> float

    pass in the reeg and return the median value

    >rEEG_median([2,3,4,5,6,9,9])
    5

    """
    return np.nanmedian(reeg)


def rEEG_lower_margin(reeg):
    """
    (list) -> float

    pass in the reeg and return the 5th percentile value

    >rEEG_lower_margin([2,3,4,5,6,9,9])
    2

    """

    return mat_percentile(reeg, 5)


def rEEG_upper_margin(reeg):
    """
    (list) -> float

    pass in the reeg and return the 95th percentile value

    >rEEG_upper_margin([2,3,4,5,6,9,9])
    9

    """
    return mat_percentile(reeg, 95)


def rEEG_width(reeg):
    """
    (list) -> float

    pass in the reeg and return the bandwidth

    >rEEG_width([2,3,4,5,6,9,9])
    7

    """
    return rEEG_upper_margin(reeg) - rEEG_lower_margin(reeg)


def rEEG_SD(reeg):
    """
    (list) -> float

    pass in the reeg and return the standard deviation

    >rEEG_SD([2,3,4,5,6,9,9])
    2.7603

    """
    return np.nanstd(reeg, ddof=1)


def rEEG_CV(reeg):
    """
    (list) -> float

    pass in the reeg and return the coefficient of variance

    >rEEG_CV([2,3,4,5,6,9,9])
    0.5085

    """
    return np.nanstd(reeg, ddof=1) / np.nanmean(reeg)


def rEEG_asymmetry(reeg):
    """
    (list) -> float

    pass in the reeg and return the asymmetry

    >rEEG_asymmetry([2,3,4,5,6,9,9])
    0.1429

    """
    A = np.nanmedian(reeg) - rEEG_lower_margin(reeg)
    B = rEEG_upper_margin(reeg) - np.nanmedian(reeg)
    return (B - A) / (A + B)


def main_rEEG(x, Fs, feat_name, params=None):
    """
    Syntax: featx = rEEG(x, Fs, feat_name, params)
    
    Inputs:
        x          - epoch of EEG data (size 1 x N)
        Fs         - sampling frequency (in Hz)
        feat_name  - feature type, defaults to 'rEEG_mean';
                     see full list of 'rEEG_' features in params['FEATURE_SET_ALL']
        params  - parameters (as dictionary);
                     see NEURAL_parameters() for examples
    
    Outputs:
        featx  - feature at each frequency band
    
    Example:
        import utils
        import rEEG
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values
    
        featx = rEEG.main_rEEG(x, Fs, 'rEEG_width')


    [1] D Oâ€™Reilly, MA Navakatikyan, M Filip, D Greene, & LJ Van Marter (2012). Peak-to-peak
    amplitude in neonatal brain monitoring of premature infants. Clinical Neurophysiology,
    123(11), 2139â€“53.
    """
    if params is None:
        params = NEURAL_parameters.NEURAL_parameters()
        if "rEEG" in params:
            params = params["rEEG"]
        else:
            raise ValueError("No default parameters found")
    elif len(params) == 0:
        params = NEURAL_parameters.NEURAL_parameters()
        if "rEEG" in params:
            params = params["rEEG"]
        else:
            raise ValueError("No default parameters found")

    freq_bands = params["freq_bands"]

    N_freq_bands = len(freq_bands)
    if N_freq_bands == 0:
        N_freq_bands = 1

    x_orig = x.copy()

    featx = np.empty(N_freq_bands)
    for n in range(N_freq_bands):
        if not len(freq_bands) == 0:
            x_filt, dum = utils.filter_butterworth_withnans(
                x_orig,
                Fs,
                freq_bands[n][1],
                freq_bands[n][0],
                5,
                params["FILTER_REPLACE_ARTEFACTS"],
            )
            x_orig = x.copy()
        else:
            x_filt = x.copy()

        reeg = gen_rEEG(
            x_filt,
            params["overlap"],
            params["L_window"],
            params["window_type"],
            Fs,
            params["APPLY_LOG_LINEAR_SCALE"],
        )

        try:
            featx[n] = eval(feat_name)(reeg)
        except:
            raise ValueError("Feature function not found: %s" % feat_name)

    return featx



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

from NEURAL_py_EEG import amplitude_features
from NEURAL_py_EEG import spectral_features
from NEURAL_py_EEG import connectivity_features
from NEURAL_py_EEG import rEEG
from NEURAL_py_EEG import fd_features
from NEURAL_py_EEG import utils
import numpy as np
import pandas as pd
from NEURAL_py_EEG import IBI_features
from NEURAL_py_EEG import NEURAL_parameters


def size_feature(feat_name):
    """
    size_feature: size of feature (e.g. array or 1 point)

    Syntax: N = size_feature(feat_name)

    Inputs:
        feat_name - feature name (e.g. 'spectral_power')

    Outputs:
        N - number of frequency bands

    Example:
        import generate_all_features
        N = generate_all_features.size_feature('spectral_power')
        print('Feature has %d frequency bands\n' % N)

    :param feat_name:  feature name (e.g. 'spectral_power')
    :return: N - number of frequency bands
    """

    params = NEURAL_parameters.NEURAL_parameters()
    feat_group = feat_name.split("_")

    feat_group = feat_group[0]

    if feat_group == "IBI":
        return 1

    if feat_name == "spectral_edge_frequency":
        # ---------------------------------------------------------------------
        # SPECIAL cases
        # ---------------------------------------------------------------------
        N = 1
    else:
        # ---------------------------------------------------------------------
        # feature set is size number of frequency bands
        # ---------------------------------------------------------------------
        bands = np.array(params[feat_group]["freq_bands"])
        if bands.ndim == 1:
            N = 1
        else:
            N = bands.shape[0]

    return N


def overlap_epochs(x, Fs, L_window, overlap=50, window_type="rect"):
    """
        ---------------------------------------------------------------------
        overlapping epochs in one matrix
        ---------------------------------------------------------------------

    :param x:
    :param Fs:
    :param L_window:
    :param overlap:
    :param window_type:
    :return:
    """

    L_hop, L_epoch, win_epoch = utils.gen_epoch_window(
        overlap, L_window, window_type, Fs
    )
    N = len(x)
    N_epochs = int(np.ceil((N - (L_epoch - L_hop)) / L_hop))
    if N_epochs < 1:
        N_epochs = 1
        print(
            "| WARNING: signal length is less than segment length (L_epoch - L_hop).\n"
        )
        print('| Adjust "EPOCH_LENGTH" or "EPOCH_OVERLAP" in ')
        print('"utils.py" file.\n')

    nw = np.array(list(range(L_epoch)))
    ix = np.array(list(range(N)))

    x_epochs = np.empty([N_epochs, L_epoch])
    x_epochs.fill(np.nan)

    for k in range(N_epochs):
        nf = nw + k * L_hop
        # zero-pad if outside x
        if np.max(nf) > ix[-1]:
            nf = np.delete(nf, np.where(nf > ix[-1])[0])
        i_nf = np.array(list(range(len(nf))))

        x_epochs[k, i_nf] = x[nf] * win_epoch[i_nf]

    return x_epochs


def generate_all_features(data, save_p_l=None, channel_names=None, feat_set=None):
    """
    Syntax: feat_st = generate_all_features(data, save_path, channel_names, feat_set)

    Inputs:
        data          - either EEG filename or data dictionary with EEG
                         e.g. data dictionary: data_st = utils.gen_test_EEGdata(5*60, 64, 1)
        channel_names  - channel labels to process (default, process all)
                         e.g. ['C3-O1','C4-O2','F3-C3','F4-C4']
        feat_set       - list of features to compute,
                         e.g. ['spectral_relative_power','rEEG_SD', 'connectivity_BSI']

    Outputs:
        feats_per_epochs - features estimated over all epochs
        feat_pd_names - feature names for dataframe
        feat_st - dictionary containing results - not separated into individual frequency bands
        feats_median_ch - DataFrame containing the the median over all channels. Epoch output. Separated in freq bands
        feats_median_all - DataFrame containing the the median over all channels and epochs. Separated in freq bands

    Example:
        import utils
        import generate_all_features

        # generate 5 minutes of simulated multichannel EEG data, with 64 Hz sample frequency
        data_st = utils.gen_test_EEGdata(5 * 60, 64, 1)

        # select features to compute:
        feature_set = ['spectral_relative_power', 'rEEG_SD', 'connectivity_BSI']

        # generate all features:
        feats_per_epochs, feat_pd_names, feat_st, feats_median_ch, feats_median_all =
        generate_all_features.generate_all_features(data_st, feat_set=feature_set)
    """

    params = NEURAL_parameters.NEURAL_parameters()

    # ---------------------------------------------------------------------
    # 1. load EEG data from .csv file
    # ---------------------------------------------------------------------

    if isinstance(data, dict):
        eeg_data = data["eeg_data"]
        Fs = data["Fs"]
        ch_labels = np.array(data["ch_labels"])
    else:
        eeg_data = pd.read_csv(data)
        Fs = int(data.split("_", -1)[-1][0:-4])
        ch_labels = np.array(eeg_data.columns)

    feats_per_epochs = np.array([])

    # select channels
    if not not channel_names:  # Check if channel_names is empty
        ikeep = []
        for n in range(len(channel_names)):
            it = [ind for ind, x in enumerate(ch_labels) if x == channel_names[n]]
            if not not it:
                ikeep.append(it)

        rem = [x for x in range(len(ch_labels)) if x not in ikeep]
        eeg_data = eeg_data.drop([ch_labels[rem]])
        ch_labels = ch_labels[ikeep]

    # or remove empty channels
    irem = []
    for n in range(len(ch_labels)):
        if all(np.isnan(eeg_data[ch_labels[n]])):
            irem.append(n)

    if not not irem:
        for r in range(len(irem)):
            eeg_data = eeg_data.drop([ch_labels[irem[r]]], axis=1)
        ch_labels = np.delete(ch_labels, irem)

    N_channels = len(eeg_data.columns)

    # ---------------------------------------------------------------------
    # 2. generate features
    # ---------------------------------------------------------------------
    if feat_set is None:
        feat_set = params["FEATURE_SET_ALL"]

    N_feats = len(feat_set)

    # A) iterate over features

    max_num_epochs = get_max_num_epochs(params, len(eeg_data), Fs)
    feat_pd_names = get_feature_names(params, feat_set)
    tmp1 = np.zeros([1, len(feat_pd_names)])
    tmp2 = np.zeros([max_num_epochs, len(feat_pd_names)])
    tmp1.fill(np.nan)
    tmp2.fill(np.nan)
    feats_median_all = pd.DataFrame(data=tmp1, columns=feat_pd_names)
    feats_median_ch = pd.DataFrame(
        data=tmp2[:, :98], columns=feat_pd_names[:98]
    )  # all names except IBI

    feat_st = dict()
    for n in range(N_feats):
        # print('%i / %i' % (n, N_feats-1))
        L_feature = size_feature(feat_set[n])
        feat_group = feat_set[n].split("_")
        feat_group = feat_group[0]

        # ---------------------------------------------------------------------
        # SPECTRAL and AMPLITUDE
        # (analysis on a per-channel basis and divide each channel into epochs)
        # ---------------------------------------------------------------------
        if feat_group in ["amplitude", "spectral", "rEEG", "FD"]:

            # B) iterate over all channels
            feats_channel = np.empty([N_channels, L_feature])
            feats_channel.fill(np.nan)

            for c in range(N_channels):
                x_epochs = overlap_epochs(
                    eeg_data[ch_labels[c]],
                    Fs,
                    params["EPOCH_LENGTH"],
                    params["EPOCH_OVERLAP"],
                )
                N_epochs, dum = x_epochs.shape

                if c == 0:
                    feats_per_epochs = np.empty([N_channels, N_epochs, L_feature])
                    feats_per_epochs.fill(np.nan)

                # C) iterate over epochs
                feats_epochs = np.empty([N_epochs, L_feature])
                feats_epochs.fill(np.nan)

                for e in range(N_epochs):
                    L_nans = len(np.where(np.isnan(x_epochs[e, :]))[0])
                    if (
                        100 * (L_nans / len(x_epochs[e, :]))
                        < params["EPOCH_IGNORE_PRC_NANS"]
                    ):
                        if feat_group == "spectral":
                            tmp = spectral_features.main_spectral(
                                x_epochs[e, :], Fs, feat_set[n]
                            )
                        elif feat_group == "FD":
                            tmp = fd_features.main_fd(x_epochs[e, :], Fs, feat_set[n])
                        elif feat_group == "amplitude":
                            tmp = amplitude_features.main_amplitude(
                                x_epochs[e, :], Fs, feat_set[n]
                            )
                        elif feat_group == "rEEG":
                            tmp = rEEG.main_rEEG(x_epochs[e, :], Fs, feat_set[n])
                        else:
                            raise ValueError(
                                'Incorrect "feat_group" - should not have entered here'
                            )

                        feats_epochs[e, :] = tmp
                feats_per_epochs[c, :, :] = feats_epochs
                # Median over all epochs
                feats_channel[c, :] = np.nanmedian(feats_epochs, axis=0)
            # Median over all channels
            feat_st[feat_set[n]] = np.nanmedian(feats_channel, axis=0)

            feats_median_ch = add_feat_data_to_array(
                feats_median_ch,
                np.median(feats_per_epochs, axis=0),
                feat_pd_names,
                feat_set[n],
                params,
            )
            feats_median_all = add_feat_data_to_array(
                feats_median_all,
                np.array([np.median(feats_channel, axis=0)]),
                feat_pd_names,
                feat_set[n],
                params,
            )
        # ---------------------------------------------------------------------
        # CONNECTIVITY FEATURES
        # (use over all channels but also divide into epochs)
        # ---------------------------------------------------------------------
        elif feat_group == "connectivity":
            dum_row = []
            x_epochs = []  # This is only to suppress warning
            for c in range(N_channels):
                tmp = overlap_epochs(
                    eeg_data[ch_labels[c]],
                    Fs,
                    params["EPOCH_LENGTH"],
                    params["EPOCH_OVERLAP"],
                )
                if c == 0:
                    dum_row, dumy = tmp.shape
                    x_epochs = np.empty([N_channels, dum_row, dumy])
                    x_epochs.fill(np.nan)
                x_epochs[c, :, :] = tmp

            N_epochs = dum_row

            # B) iterate over epochs:
            feats_epochs = np.empty([N_epochs, L_feature])
            feats_epochs.fill(np.nan)
            for e in range(N_epochs):
                x_ep = x_epochs[:, e, :]
                L_nans = len(np.where(np.isnan(x_ep))[0])
                if 100 * (L_nans / x_ep.size) < params["EPOCH_IGNORE_PRC_NANS"]:
                    tmp = connectivity_features.main_connectivity(
                        x_ep, Fs, feat_set[n], ch_labels
                    )
                    feats_epochs[e, :] = tmp
            # Median over all epochs
            feat_st[feat_set[n]] = np.nanmedian(feats_epochs, axis=0)
            feats_median_ch = add_feat_data_to_array(
                feats_median_ch, feats_epochs, feat_pd_names, feat_set[n], params
            )
            feats_median_all = add_feat_data_to_array(
                feats_median_all,
                np.array([np.nanmedian(feats_epochs, axis=0)]),
                feat_pd_names,
                feat_set[n],
                params,
            )

        elif feat_group == "IBI":
            # B) iterate over epochs:
            feats_channel = np.empty([N_channels, L_feature])
            feats_channel.fill(np.nan)
            for c in range(N_channels):
                feats_channel[c, :] = IBI_features.main_IBI(
                    eeg_data[ch_labels[c]].to_numpy(), Fs, feat_set[n]
                )

            # and the median over all channels
            feat_st[feat_set[n]] = np.nanmedian(feats_channel, axis=0)
            feats_median_all = add_feat_data_to_array(
                feats_median_all,
                np.array([np.nanmedian(feats_channel, axis=0)]),
                feat_pd_names,
                feat_set[n],
                params,
            )

    if save_p_l is not None:
        save_p_l = save_p_l + "_all_med"
        utils.save_pd_as_file(feats_median_all, save_p_l)
        save_p_l = save_p_l + "_ch_med"
        utils.save_pd_as_file(feats_median_ch, save_p_l)

    return feats_per_epochs, feat_pd_names, feat_st, feats_median_ch, feats_median_all


def get_feature_names(params, feat_set):
    """
    The purpose of this function to create an array that contains all the features names. This is the feature name,
    channel and frequency band. Array will be nans and one for every channel in the 'BI_MONT' param except for the
    connectivity features

    :param feat_set: The feature set being examined
    :param params: The parameters for the neural dataset
    :return: numpy string array containing feature names for all different frequency bands
    """
    all_feats = params["FEATURE_SET_ALL"]
    names = []
    for feat in all_feats:
        if feat not in feat_set:
            continue
        feat_group = feat.split("_")
        feat_group = feat_group[0]

        if feat_group == "IBI":
            bands = np.array([0.5, 30])
        elif (
            "total_freq_bands" in params[feat_group].keys()
            and feat == "spectral_edge_frequency"
        ):
            bands = np.array(params[feat_group]["total_freq_bands"])
        else:
            bands = np.array(params[feat_group]["freq_bands"])

        if bands.ndim == 1:
            N_bands = 1
        else:
            N_bands = bands.shape[0]

        for b in range(N_bands):
            if N_bands == 1:
                b1 = bands[0]
                b2 = str(int(bands[1]))
            else:
                b1 = bands[b][0]
                b2 = str(int(bands[b][1]))

            if b1 != 0.5:
                b1 = str(int(b1))
            else:
                b1 = str(b1).replace(".", "")

            names.append(feat + "__" + b1 + "_" + b2)

    return names


def get_max_num_epochs(params, N, Fs):
    """
    The purpose of this function is to find what the maximum number of epochs is so an empty nan np array can be made

        Inputs:
        params:           the parameter set
        N:                the number of samples in a single channel of an EEG
        Fs:               sampling frequency

    Outputs:
        a:                returns the number of epochs that will be generated

    NOTE: If different window sizes or overlaps are created then this function should be updated

    """
    # num_epochs = []
    L_window = params["EPOCH_LENGTH"]
    overlap = params["EPOCH_OVERLAP"]

    L_hop, L_epoch, dum = utils.gen_epoch_window(overlap, L_window, "hamm", Fs)
    a = np.ceil((N - (L_epoch - L_hop)) / L_hop)

    return int(a)


def add_feat_data_to_array(all_data, new_data, feat_pd_names, feat_name, params):
    """
    Update all_data array by appending the new data. The new data contains the data from all freq bands

    Syntax: all_data = add_feat_data_to_array(all_data, new_data, feat_pd_names, feat_name, params)

    Inputs:
        all_data:               Full data array for all features
        new_data:               New data for all feat bands of 'feat_name'
        feat_pd_names:          All the feature names (with bands and channels)
        feat_name:              Name of current feature
        params:                 Needed to get the feature group

    Outputs:
        packaged_data:    dictionary including EEG data (referential montage), sampling
                          frequency, and channel labels. Data are pandas DataFrame
    """

    feat_group = feat_name.split("_")
    feat_group = feat_group[0]
    if feat_group == "IBI":
        bands = np.array([0.5, 30])
    elif (
        "total_freq_bands" in params[feat_group].keys()
        and feat_name == "spectral_edge_frequency"
    ):
        bands = np.array(params[feat_group]["total_freq_bands"])
    else:
        bands = np.array(params[feat_group]["freq_bands"])

    if bands.ndim == 1:
        N_bands = 1
    else:
        N_bands = bands.shape[0]

    for b in range(N_bands):

        if N_bands == 1:
            b1 = bands[0]
            b2 = str(int(bands[1]))
        else:
            b1 = bands[b][0]
            b2 = str(int(bands[b][1]))

        if b1 != 0.5:
            b1 = str(int(b1))
        else:
            b1 = str(b1).replace(".", "")

        a_feat_name = feat_name + "__" + b1 + "_" + b2

        if a_feat_name not in feat_pd_names:
            raise ValueError(
                "Feature name %s does not exist - check code" % a_feat_name
            )

        # idx = [i for i, s in enumerate(feat_pd_names) if s == a_feat_name]

        # print(a_feat_name)
        if new_data.shape[0] == 1:
            all_data[a_feat_name] = [new_data[0][b]]
        else:
            all_data[a_feat_name] = new_data[:, b]

    return all_data



# Author:       Brian Murphy
# Date started: 17/02/2020
# Last updated: <17/02/2020 12:42:22 (BrianM)>

def NEURAL_parameters():
    """
    This function is to generate all the default NEURAL parameters

    :return: a dictionary containing all the system parameters
    """
    params = dict()

    '''
    ---------------------------------------------------------------------
    Proprocessing (lowpass filter and resample)
    ---------------------------------------------------------------------
    '''
    params['LP_fc'] = 30  # low pass filter cut-off
    params['Fs_new'] = 64  # down_sample to Fs_new

    '''
    ---------------------------------------------------------------------
    Directories - fill in directories
    ---------------------------------------------------------------------
    '''
    params['EEG_DATA_DIR'] = ''
    params['EEG_DATA_DIR_CSV'] = ''

    '''
    ---------------------------------------------------------------------
    Montage - bipolar mantage of NICU babies
    ---------------------------------------------------------------------
    '''
    params['BI_MONT'] = [['F4', 'C4'], ['F3', 'C3'], ['C4', 'T4'], ['C3', 'T3'], ['C4', 'Cz'], ['Cz', 'C3'],
                         ['C4', 'O2'], ['C3', 'O1']]

    '''
    ---------------------------------------------------------------------
    Artefacts
    ---------------------------------------------------------------------
    '''
    params['REMOVE_ART'] = 1  # simple procedure to remove artefacts; 0 to turn off

    # some default values used for preterm infants (<32 weeks of gestation)
    params['ART_HIGH_VOLT'] = 1500  # in mirco Vs
    params['ART_TIME_COLLAR'] = 10  # time collar( in seconds) around high - amplitude artefact

    params['ART_DIFF_VOLT'] = 200  # in mirco Vs
    params['ART_DIFF_TIME_COLLAR'] = 0.5  # time collar( in seconds) around fast jumps
    params['ART_DIFF_MIN_TIME'] = 0.1  # min time( in seconds) for flat(continuous) trace to be artefact

    params['ART_ELEC_CHECK'] = 1  # minimum length required for electrode check( in seconds)

    params['ART_REF_LOW_CORR'] = 0.15  # if mean correlation coefficent across referential channels
    # is < this value then remove

    # what to replace artefacts with before filtering?
    # options: 1) zeros('zeros')
    #          2) linear interpolation('linear_interp')
    #          3) cubic spline interpolation('cubic_interp')
    #          4) NaN('nans'): replace with cubic spline before filtering and then NaN's % after filtering
    params['FILTER_REPLACE_ARTEFACTS'] = 'nans'

    params['amplitude'] = dict()
    params['rEEG'] = dict()
    params['connectivity'] = dict()
    params['FD'] = dict()

    params['amplitude']['FILTER_REPLACE_ARTEFACTS'] = params['FILTER_REPLACE_ARTEFACTS']
    params['rEEG']['FILTER_REPLACE_ARTEFACTS'] = params['FILTER_REPLACE_ARTEFACTS']
    params['connectivity']['FILTER_REPLACE_ARTEFACTS'] = params['FILTER_REPLACE_ARTEFACTS']
    params['FD']['FILTER_REPLACE_ARTEFACTS'] = params['FILTER_REPLACE_ARTEFACTS']

    '''
    ---------------------------------------------------------------------
    Features
    ---------------------------------------------------------------------
    '''
    params['FEATURE_SET_ALL'] = [
        'spectral_power'
        , 'spectral_relative_power'
        , 'spectral_flatness'
        , 'spectral_diff'
        , 'spectral_entropy'
        , 'spectral_edge_frequency'
        , 'FD'
        , 'amplitude_total_power'
        , 'amplitude_SD'
        , 'amplitude_skew'
        , 'amplitude_kurtosis'
        , 'amplitude_env_mean'
        , 'amplitude_env_SD'
        , 'connectivity_BSI'
        , 'connectivity_corr'
        , 'connectivity_coh_mean'
        , 'connectivity_coh_max'
        , 'connectivity_coh_freqmax'
        , 'rEEG_mean'
        , 'rEEG_median'
        , 'rEEG_lower_margin'
        , 'rEEG_upper_margin'
        , 'rEEG_width'
        , 'rEEG_SD'
        , 'rEEG_CV'
        , 'rEEG_asymmetry'
        , 'IBI_length_max'
        , 'IBI_length_median'
        , 'IBI_burst_prc'
        , 'IBI_burst_number'
    ]

    '''
    Frequency bands
    '''
    params['FREQ_BANDS'] = [[0.5, 4], [4, 7], [7, 13], [13, 30]]

    # these bands often used for preterm infants( < 32 weeks GA):
    # params['FREQ_BANDS'] = [[0.5, 3], [3, 8], [8, 15], [15, 30]];

    '''
    ---------------------------------------------------------------------
    A.spectral features
    ---------------------------------------------------------------------
    '''

    # how to estimate the spectrum for 'spectral_flatness', 'spectral_entropy', % spectral_edge_frequency features:
    #    1) PSD: estimate power spectral density(e.g.Welch periodogram)
    #    2) robust - PSD: median(instead of mean) of spectrogram
    #    3) periodogram: magnitude of the discrete Fourier transform
    params['spectral'] = dict()
    params['spectral']['method'] = 'PSD'

    # length of time - domain analysis window and overlap:
    # (applies to 'spectral_power', 'spectral_relative_power',
    # 'spectral_flatness', and 'spectral_diff' features)
    params['spectral']['L_window'] = 2  # in seconds
    params['spectral']['window_type'] = 'hamm'  # type of window
    params['spectral']['overlap'] = 50  # overlap in percentage
    params['spectral']['freq_bands'] = params['FREQ_BANDS']
    params['spectral']['total_freq_bands'] = [params['FREQ_BANDS'][0][0], params['FREQ_BANDS'][-1][-1]]
    params['spectral']['SEF'] = 0.95  # spectral edge frequency

    # fractal dimension(FD):
    params['FD']['method'] = 'higuchi'  # method to estimate FD, either 'higuchi' or 'katz'
    params['FD']['freq_bands'] = [params['FREQ_BANDS'][0][0], params['FREQ_BANDS'][-1][-1]]
    # $$$ params['FD']['freq_bands'] = params['FREQ_BANDS']
    params['FD']['qmax'] = 6  # Higuchi method: max.value of k

    '''
    ---------------------------------------------------------------------
    B. amplitude features
    ---------------------------------------------------------------------
    '''

    # $$$ params['amplitude']['freq_bands'] = [params['FREQ_BANDS'][0][0], params['FREQ_BANDS'][-1][-1]]
    params['amplitude']['freq_bands'] = params['FREQ_BANDS']

    # for rEEG(range - EEG, similar to aEEG) from [1]
    #
    # [1] DOâ€™Reilly, MA Navakatikyan, M Filip, D Greene, & LJ Van Marter(2012).Peak - to - peak amplitude in neonatal
    # brain monitoring of premature infants.Clinical Neurophysiology, 123(11), 2139 â€“53.
    #
    # settings in [1]: window = 2 seconds; overlap = 0 %; and no log - linear scale
    params['rEEG']['L_window'] = 2  # in seconds
    params['rEEG']['window_type'] = 'rect'  # type of window
    params['rEEG']['overlap'] = 0  # overlap in percentage
    params['rEEG']['APPLY_LOG_LINEAR_SCALE'] = 0  # use this scale(either 0 or 1)
    params['rEEG']['freq_bands'] = params['FREQ_BANDS']

    '''
    ---------------------------------------------------------------------
    C. connectivity features
    ---------------------------------------------------------------------
    '''

    # how to estimate the cross spectrum for the coherence function:
    # 1) PSD: estimate power spectral density (e.g. Welch periodogram)
    # 2) bartlett-PSD: Welch periodogram with 0% overlap and rectangular window
    #    (necessary if using the analytic assessment of zero coherence, see below)
    params['connectivity']['method'] = 'bartlett-PSD'

    params['connectivity']['freq_bands'] = params['FREQ_BANDS']
    params['connectivity']['L_window'] = 8  # PSD window in seconds
    params['connectivity']['overlap'] = 75  # PSD window percentage overlap
    params['connectivity']['window_type'] = 'hamm'  # PSD window type

    # find lower coherence limit using either either a surrogate-data
    # approach [1] or an analytic threshold [2]
    # [1] Faes L, Pinna GD, Porta A, Maestri R, Nollo G (2004). Surrogate data analysis for
    #     assessing the significance of the coherence function. IEEE Transactions on
    #     Biomedical Engineering, 51(7):1156â€“1166.
    # [2] Halliday, DM, Rosenberg, JR, Amjad, AM, Breeze, P, Conway, BA, &
    #     Farmer, SF. (1995). A framework for the analysis of mixed time series/point
    #     process data--theory and application to the study of physiological tremor, single
    #     motor unit discharges and electromyograms. Progress in Biophysics and Molecular
    #     Biology, 64(2â€“3), 237â€“278.
    #
    # options for 'feat_params_st.connectivity.coherence_zero_level' are:
    # 1) 'surr' for [1]
    # 2) 'analytic' for [2]
    # 3) '' not to implement (no threshold)
    params['connectivity']['coherence_zero_level'] = 'analytic'
    # alpha value for null-hypothesis disribution cut-off:
    params['connectivity']['coherence_zero_alpha'] = 0.05
    # number of iterations required to generate null-hypothesis distribution if
    # using surrogate data approach ([2]):
    params['connectivity']['coherence_surr_iter'] = 500

    '''
    ---------------------------------------------------------------------
    Short-time analysis on EEG
    ---------------------------------------------------------------------
    '''

    params['EPOCH_LENGTH'] = 64  # seconds
    params['EPOCH_OVERLAP'] = 50  # percent

    params['EPOCH_IGNORE_PRC_NANS'] = 50  # if epoch has â‰¥ EPOCH_IGNORE_PRC_NANS (percent) then ignore

    return params



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

import numpy as np
from NEURAL_py_EEG import utils
from NEURAL_py_EEG import NEURAL_parameters


def fd_higuchi(x, kmax="Nope"):
    """
    ---------------------------------------------------------------------
     Higuchi estimate in [1]
    ---------------------------------------------------------------------
    """
    N = len(x)
    if kmax == "Nope":
        kmax = int(np.floor(len(x) / 10))

    # what value of k to compute
    ik = 1
    k_all = np.empty(0)
    knew = 0

    while knew < kmax:
        if ik <= 4:
            knew = ik
        else:
            knew = np.floor(2 ** ((ik + 5) / 4))

        if knew <= kmax:
            k_all = np.append(k_all, knew)
        ik = ik + 1

    """
    ---------------------------------------------------------------------
    curve length for each vector:
    ---------------------------------------------------------------------
    """

    inext = 0
    L_avg = np.zeros(len(k_all))

    for k in k_all:
        L = np.zeros(k.astype(int))
        for m in range(k.astype(int)):
            ik = np.array(range(1, np.floor((N - m - 1) / k).astype(int) + 1))
            scale_factor = (N - 1) / (np.floor((N - m - 1) / k) * k)
            L[m] = np.nansum(
                np.abs(
                    x[m + np.array(ik) * k.astype(int)]
                    - x[m + (ik - 1) * k.astype(int)]
                )
            ) * (scale_factor / k)

        L_avg[inext] = np.nanmean(L)
        inext += 1

    """
    -------------------------------------------------------------------
     form log-log plot of scale v. curve length
    -------------------------------------------------------------------
    """

    x1 = np.log2(k_all)
    y1 = np.log2(L_avg)
    c = np.polyfit(x1, y1, 1)
    FD = -c[0]

    # y_fit = c[0]*x1 + c[1]
    # y_residuals = y1 - y_fit
    # r2 = 1 - np.sum(y_residuals**2) / ((N-1) * np.var(y1))

    # return FD, r2, k_all, L_avg
    return FD


def fd_katz(x, dum=0):

    """
    ---------------------------------------------------------------------
    Katz estimate in [2]
    ---------------------------------------------------------------------
    """
    N = len(x)
    p = N - 1

    # 1. line-length
    L = np.empty(N - 1)
    for n in range(N - 1):
        L[n] = np.sqrt(1 + (x[n] - x[n + 1]) ** 2)

    L = np.sum(L)

    # 2. maximum distance
    d = np.zeros(p)
    for n in range(N - 1):
        d[n] = np.sqrt((n + 1) ** 2 + (x[0] - x[n + 1]) ** 2)

    d = np.max(d)

    D = np.log(p) / (np.log(d / L) + np.log(p))

    return D


def main_fd(x, fs, feat_name="fd_higuchi", params=None):
    """
    Syntax: featx = fd_features(x, Fs, params)

    Inputs:
        x          - epoch of EEG data (size 1 x N)
        Fs         - sampling frequency (in Hz)
        params     - parameters (as dictionary)
                     see NEURAL_parameters() for examples

    Outputs:
        featx  - fractal dimension for each frequency band

    Example:
        import utils
        import fd_features
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values

        featx = fd_features.main_fd(x, Fs)

    [1] T Higuchi, â€œApproach to an irregular time series on the basis of the fractal
    theory,â€� Phys. D Nonlinear Phenom., vol. 31, pp. 277â€“283, 1988.

    [2] MJ Katz, Fractals and the analysis of waveforms. Computers in Biology and Medicine,
    vol. 18, no. 3, pp. 145â€“156. 1988
    """
    if feat_name == "FD":
        feat_name = "fd_higuchi"

    if params is None:
        params = NEURAL_parameters.NEURAL_parameters()
        if "FD" in params:
            params = params["FD"]
        else:
            raise ValueError("No default parameters found")
    elif len(params) == 0:
        params = NEURAL_parameters.NEURAL_parameters()
        if "FD" in params:
            params = params["FD"]
        else:
            raise ValueError("No default parameters found")

    freq_bands = np.array(params["freq_bands"])

    N_freq_bands = freq_bands.ndim
    if N_freq_bands == 0:
        N_freq_bands = 1

    x_orig = x.copy()

    featx = np.empty(N_freq_bands)
    for n in range(N_freq_bands):
        if N_freq_bands == 1:
            x, dum = utils.filter_butterworth_withnans(
                x_orig,
                fs,
                freq_bands[1],
                freq_bands[0],
                5,
                params["FILTER_REPLACE_ARTEFACTS"],
            )
        else:
            x, dum = utils.filter_butterworth_withnans(
                x_orig,
                fs,
                freq_bands[n][1],
                freq_bands[n][0],
                5,
                params["FILTER_REPLACE_ARTEFACTS"],
            )

        try:
            featx[n] = eval(feat_name)(x, params["qmax"])
        except:
            raise ValueError("Feature function not found: %s" % feat_name)

    return featx



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <14/02/2020 14:57:19 (BrianM)>

"""
Python implementation of Matlab's 'filtfilt' function
Example:
    x = np.random.randn(1001, )
    b, a = signal.butter(1, 0.5 / 32, 'high')
    y = mat_filtfilt(b, a, x)


John M. O' Toole, University College Cork
Started: 06-11-2019
last update: Time-stamp: <2019-11-06 16:20:45 (otoolej)>
"""
import numpy as np
from scipy import sparse, signal


def gain_filt(b, a):
    """Step response from filter (used as initial conditions later)

    Parameters
    ----------
    b: ndarray
        filter coefficients
    a: ndarray
        filter coefficients

    Returns
    -------
    zi : ndarray
        step response as a vector
    """
    # max. length of coefficients:
    if len(a) == 1 and a[0] == 1:
        a = np.ones([len(b)])
    N_filt = max(len(b), len(a))

    # arrange entries of sparse matrix:
    rows = [*range(N_filt - 1), *range(1, N_filt - 1), *range(N_filt - 2)]
    cols = [*np.zeros(N_filt - 1).astype(int), *
            range(1, N_filt - 1), *range(1, N_filt - 1)]
    vals = [*np.hstack(((1 + a[1]), a[2:N_filt])),
            *np.ones(N_filt - 2).astype(int),
            *-np.ones(N_filt - 2).astype(int)]
    rhs = b[1:N_filt] - b[0] * a[1:N_filt]

    AA = sparse.coo_matrix((vals, (rows, cols))).tocsr()
    zi = sparse.linalg.spsolve(AA, rhs)

    return (zi)



def mat_filtfilt(b, a, x):
    """Fowards--backwards filter to match Matlab's 'filtfilt' function


    Parameters
    ----------
    b: ndarray
        filter coefficients
    a: ndarray
        filter coefficients
    x: ndarray
        input signal

    Returns
    -------
    y : ndarray
        filtered signal
    """
    # 1. pad the signal:
    L_pad = 3 * (max(len(b), len(a)) - 1)
    x_pad = np.concatenate((2 * x[0] - x[1:(L_pad + 1)][::-1],
                            x,
                            2 * x[-1] - x[len(x)-L_pad-1:-1][::-1]))

    # 2. estimate initial filter conditions:
    zi = gain_filt(b, a)

    # 3. forwards and backwards filter:
    x_pad, _ = signal.lfilter(b, a, x_pad, zi=zi * x_pad[0])
    x_pad = x_pad[::-1]
    x_pad, _ = signal.lfilter(b, a, x_pad, zi=zi * x_pad[0])
    x_pad = x_pad[::-1]

    # 4. remove the padding:
    y = x_pad[L_pad:len(x) + L_pad]

    return (y)



# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

import numpy as np
from scipy import signal
from scipy import stats
from NEURAL_py_EEG import utils
from NEURAL_py_EEG import NEURAL_parameters


def amplitude_total_power(data):
    """
    (list) -> float

    pass in an array or list and return the signal power

    >signal_power([3,7,8,9,1])
    40.8
    """
    return np.nanmean(np.abs(data) ** 2)


def amplitude_env_mean(data):
    """
    (list) -> float

    pass in an array and return the mean value of the envelope

    >signal_envelope([3,7,8,9,1])
    50.24

    """

    return np.nanmean(np.abs(signal.hilbert(data)) ** 2)


def amplitude_SD(data):
    """
    (list) -> float

    pass in an array or list and return the standard deviation

    >signal_sd([3,7,8,9,1])
    3.4351
    """
    return np.nanstd(data, ddof=1)


def amplitude_env_SD(data):
    """
    (list) -> float

    pass in an array and return the standard deviation of the envelope

    >signal_envelope_sd([3,7,8,9,1])
    32.8035

    """

    return np.nanstd(np.abs(signal.hilbert(data)) ** 2, ddof=1)


def amplitude_skew(data):
    """
    (list) -> float

    pass in an array or list and return the absolute signal skewness

    >signal_skewness([3,7,8,9,1])
    0.4071
    """
    return np.abs(stats.skew(data, nan_policy="omit"))


def amplitude_kurtosis(data):
    """
    (list) -> float

    pass in an array or list and return the kurtosis value of data. Using Pearson's definition

    >signal_skewness([3,7,8,9,1])
    1.4904
    """
    return stats.kurtosis(data, fisher=False)


def main_amplitude(x, Fs, feat_name, params=None):
    """
    Syntax: featx = amplitude_features(x,Fs,feat_name,params)

    Inputs:
        x          - epoch of EEG data (size 1 x N)
        Fs         - sampling frequency (in Hz)
        feat_name  - feature type, defaults to 'amplitude_total_power';
                     see full list of 'amplitude_' features in params['FEATURE_SET_ALL']
        params     - parameters (as dictionary);

    Outputs:
        featx  - feature at each frequency band

    Example:
        import utils
        import amplitude_features
        
        Fs = 64
        data_st = utils.gen_test_EEGdata(32, Fs, 1)
        x = data_st['eeg_data'][data_st['ch_labels'][0]].values

        featx = amplitude_features.main_amplitude(x, Fs, 'amplitude_env_mean')
    """
    if params is None:
        params = NEURAL_parameters.NEURAL_parameters()
        if "amplitude" in params:
            params = params["amplitude"]
        else:
            raise ValueError("No default parameters found")
    elif len(params) == 0:
        params = NEURAL_parameters.NEURAL_parameters()
        if "amplitude" in params:
            params = params["amplitude"]
        else:
            raise ValueError("No default parameters found")

    freq_bands = params["freq_bands"]

    N_freq_bands = len(freq_bands)
    if N_freq_bands == 0:
        N_freq_bands = 1

    x_orig = x.copy()

    featx = np.empty(N_freq_bands)
    for n in range(N_freq_bands):
        if not len(freq_bands) == 0:
            x_filt, dum = utils.filter_butterworth_withnans(
                x_orig,
                Fs,
                freq_bands[n][1],
                freq_bands[n][0],
                5,
                params["FILTER_REPLACE_ARTEFACTS"],
            )
            x_orig = x.copy()
        else:
            x_filt = x.copy()

        try:
            x_filt = np.delete(x_filt, np.where(np.isnan(x_filt))[0])
            featx[n] = eval(feat_name)(x_filt)
        except:
            raise ValueError("Feature function not found: %s" % feat_name)

    return featx



# Author:       Brian Murphy
# Date started: 05/02/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

from NEURAL_py_EEG import utils
from NEURAL_py_EEG import preprocessing_EEG
import numpy as np
from NEURAL_py_EEG import generate_all_features
import time


def artefact_removal_examples():
    print("Running artefact_removal_examples function")

    Fs = 256  # sampling frequency
    print("Generating EEG data using the utils.gen_test_EEGdata function")
    data_st = utils.gen_test_EEGdata(2 * 60, Fs, 1)
    N = len(data_st["eeg_data_ref"])

    # simulate artefacts:
    # 1. F3 not properly attached:
    print("Simulating artefacts:")
    print("1. F3 not properly attached")
    if3 = [i for i, x in enumerate(data_st["ch_labels_ref"]) if x == "F3"][0]
    data_st["eeg_data_ref"][data_st["ch_labels_ref"][if3]] = np.random.randn(N) * 10

    # 2. electrode coupling between C4 and Cz
    print("2. Electrode coupling between C4 and Cz")
    ic4 = [i for i, x in enumerate(data_st["ch_labels_ref"]) if x == "C4"][0]
    icz = [i for i, x in enumerate(data_st["ch_labels_ref"]) if x == "Cz"][0]

    data_st["eeg_data_ref"][data_st["ch_labels_ref"][icz]] = (
        data_st["eeg_data_ref"][data_st["ch_labels_ref"][ic4]] + np.random.randn(N) * 5
    )

    # re-generate bipolar montage:
    data_st["eeg_data"], data_st["ch_labels"] = utils.set_bi_montage(
        data_st["eeg_data_ref"], Fs, data_st["ch_labels_bi"]
    )

    # remove channels:
    print("Running artefact removal")
    eeg_art = preprocessing_EEG.remove_artefacts(
        data_st["eeg_data"].copy(),
        data_st["ch_labels"],
        data_st["Fs"],
        data_st["eeg_data_ref"],
        data_st["ch_labels_ref"],
    )[0]

    print(data_st["eeg_data"].head())
    print(eeg_art.head())
    print("The artefact_removal_examples function has finished")
    print("\n\n\n\n")


def generate_a_subset_of_features_example():
    print("Running generate_a_subset_of_features_example function")
    Fs = 64  # sampling frequency

    # generate EEG-like data (coloured Gaussian noise)
    print("Generating EEG data using the utils.gen_test_EEGdata function")
    data_st = utils.gen_test_EEGdata(5 * 60, Fs, 1)

    # define feature set (or can define in utils.NEURAL_parameters):
    feature_set = ["spectral_relative_power", "rEEG_SD", "connectivity_BSI"]
    print("Feature set: %s" % feature_set)

    # estimate features:
    print("Estimating subset of features")
    (
        feats_per_epochs,
        feat_pd_names,
        feat_st,
        feats_median_ch,
        feats_median_all,
    ) = generate_all_features.generate_all_features(data_st, feat_set=feature_set)
    print("Features extracted")
    print(feat_pd_names)
    print("The generate_a_subset_of_features_example function has finsihed")
    print("\n\n\n\n")


def generate_all_features_example():
    print("Running generate_a_subset_of_features_example function")
    Fs = 64  # sampling frequency

    # generate EEG-like data (coloured Gaussian noise)
    print("Generating EEG data using the utils.gen_test_EEGdata function")
    data_st = utils.gen_test_EEGdata(5 * 60, Fs, 1)

    # estimate features:
    print("Estimating features")
    (
        feats_per_epochs,
        feat_pd_names,
        feat_st,
        feats_median_ch,
        feats_median_all,
    ) = generate_all_features.generate_all_features(data_st)

    print("Features extracted")
    print(feat_pd_names)
    print("The generate_a_subset_of_features_example function has finished")
    print("\n\n\n\n")


def preprocessing_and_feature_extraction():
    print("Running preprocessing_and_feature_extraction function")
    Fs = 256  # sampling frequency
    # generate EEG-like data (coloured Gaussian noise)
    print("Generating EEG data using the utils.gen_test_EEGdata function")
    data_st = utils.gen_test_EEGdata(5 * 60, Fs, 1)

    print("Running artefact removal")
    eeg_art = preprocessing_EEG.remove_artefacts(
        data_st["eeg_data"].copy(),
        data_st["ch_labels"],
        data_st["Fs"],
        data_st["eeg_data_ref"],
        data_st["ch_labels_ref"],
    )[0]

    # replace the montage with the artifact free version
    data_st["eeg_data"] = eeg_art.copy()

    # filter and downsample the signal
    print("Before filtering and downsampling")
    print(data_st["eeg_data"].head())
    print("Number of rows: %s" % str(data_st["eeg_data"].shape[0]))

    data_st["eeg_data"] = preprocessing_EEG.LPF_zero_phase(
        data_st["eeg_data"].copy(), data_st["Fs"]
    )
    data_st["eeg_data"], data_st["Fs"] = preprocessing_EEG.signal_downsample(
        data_st["eeg_data"].copy(), data_st["Fs"], 64
    )

    print("After filtering and downsampling")
    print(data_st["eeg_data"].head())
    print("Number of rows: %s" % str(data_st["eeg_data"].shape[0]))

    # estimate features:
    print("Estimating features")
    (
        feats_per_epochs,
        feat_pd_names,
        feat_st,
        feats_median_ch,
        feats_median_all,
    ) = generate_all_features.generate_all_features(data_st)

    print("Features extracted")
    print(feat_pd_names)
    print("The preprocessing_and_feature_extraction function has finished")
    print("\n\n\n\n")


def load_edf_preprocessing_and_feature_extraction():
    print("Running load_edf_preprocessing_and_feature_extraction function")
    Fs = 256  # sampling frequency
    print("Generating EEG data using the utils.gen_test_EEGdata function")
    data_st = utils.gen_test_EEGdata(
        10 * 60, Fs, 1
    )  # 10 minutes of fake data - need to save as edf
    eeg_ref = data_st["eeg_data_ref"]

    print('Saving EEG data as "demo_data.edf" file')
    utils.save_pd_as_edf(
        eeg_ref, "demo_data.edf", Fs=Fs
    )  # Save the fake data as edf to simulate opening a real edf

    """
    preprocessing_EEG.main_preprocessing does the following operations to the edf file
                        a) read in EEG from .edf files
                        b) remove artefacts
                        c) band-pass filter
                        d) downsample
                        e) save as .csv file (optional by setting save_converted='name', save=1)
    There are 2 outputs:    
                        0) the data dictionary containing the bipolar EEG montage, sampling frequency and channel labels
                        1) the amount of data removed during the artifact removal stage
    """

    print(
        'Loading "demo_data.edf", performing artefact removal, low pass filtering and down-sampling'
    )
    data_st = preprocessing_EEG.main_preprocessing("demo_data.edf", Fs_new=64)[0]
    print("EDF file loaded and preprocessed")

    # estimate features:
    print("Extracting features")
    t = time.time()
    (
        feats_per_epochs,
        feat_pd_names,
        feat_st,
        feats_median_ch,
        feats_median_all,
    ) = generate_all_features.generate_all_features(data_st)

    print("Features extracted - time taken %s" % str(time.time() - t))
    print(feat_pd_names)
    print("The load_edf_preprocessing_and_feature_extraction function has finished")
    print("\n\n\n\n")




# Author:       Brian Murphy
# Date started: 30/01/2020
# Last updated: <25/07/2022 16:40:13 (BrianM)>

import pandas as pd
from NEURAL_py_EEG import utils
import numpy as np
from NEURAL_py_EEG import NEURAL_parameters
from NEURAL_py_EEG.connectivity_features import channel_hemispheres
from scipy import signal
import mne


def load_edf_EEG_file(file_name, channels_to_look=None):
    """
    The purpose of this function is to read in an edf file and return a pandas dataframe containing
    the raw edf data and the sampling frequencies from all channels

    Syntax: data_out, ch_fs = load_edf_EEG_file(file_name, channels_to_look=None)

    Inputs:
        file_name:                    duration of EEG-like data in seconds (default 300 seconds)
        channels_to_look:             sampling frequency

    Outputs:
        data_out:                     pandas DataFrame containing the EEG data
        ch_fs:                        sampling frequency

    Example:
        import preprocessing_EEG
        import utils
        Fs = 256
        data_st = utils.gen_test_EEGdata(10 * 60, Fs, 1)  # 10 minutes of fake data - need to save as edf
        eeg_ref = data_st['eeg_data_ref']
        utils.save_pd_as_edf(eeg_ref, 'demo_data.edf', Fs=Fs)  # Save the fake data as edf so we can opening a real edf

        data_out, ch_fs = preprocessing_EEG.load_edf_EEG_file('demo_data.edf', channels_to_look=None)


    """
    if channels_to_look is None:
        channels_to_look = ["F3", "F4", "C3", "C4", "O1", "O2", "T3", "T4", "Cz"]
    f = mne.io.read_raw_edf(file_name, preload=True)
    data = f._data
    signal_labels = f.ch_names

    ch_names = []
    ch_fs = []
    data_out = pd.DataFrame()
    loop_index = 0

    for ch in channels_to_look:
        ch_exist = [x for x in signal_labels if ch.upper() in x.upper()]
        if len(ch_exist) > 0:
            loc = signal_labels.index(ch_exist[0])
            # data_out.insert(loop_index, ch, f.readSignal(loc))
            data_out.insert(loop_index, ch, data[loc, :] * 10 ** 6)
            ch_names.append(ch)
            tmp = f.info["sfreq"]  # [ch, f.getSampleFrequency(loc)]
            ch_fs.append(tmp)
            loop_index += 1

    # f.close
    return data_out, ch_fs


def overlap_epochs(x, Fs, L_window, overlap=50, window_type="rect"):
    """
        ---------------------------------------------------------------------
        overlapping epochs in one matrix
        ---------------------------------------------------------------------

    :param x:
    :param Fs:
    :param L_window:
    :param overlap:
    :param window_type:
    :return:
    """

    L_hop, L_epoch, win_epoch = utils.gen_epoch_window(
        overlap, L_window, window_type, Fs
    )
    N = len(x)
    N_epochs = int(np.ceil((N - (L_epoch - L_hop)) / L_hop))
    if N_epochs < 1:
        N_epochs = 1
        print(
            "| WARNING: signal length is less than segment length (L_epoch - L_hop).\n"
        )
        print('| Adjust "EPOCH_LENGTH" or "EPOCH_OVERLAP" in ')
        print('"NEURAL_parameters.py" file.\n')

    nw = np.array(list(range(L_epoch)))
    ix = np.array(list(range(N)))

    x_epochs = np.empty([N_epochs, L_epoch])
    x_epochs.fill(np.nan)
    x_epochs_inds = np.zeros([N_epochs, L_epoch])

    for k in range(N_epochs):
        nf = nw + k * L_hop
        # zero-pad if outside x
        if np.max(nf) > ix[-1]:
            nf = np.delete(nf, np.where(nf > ix[-1])[0])
        i_nf = np.array(list(range(len(nf))))

        x_epochs[k, i_nf] = x[nf] * win_epoch[i_nf]
        x_epochs_inds[k, i_nf] = nf

    return x_epochs, x_epochs_inds.astype(int)


def art_per_channel(x, Fs, DBverbose):
    """
    Remove artefacts on a per-channel basis

    :param DBverbose:
    :param x: data from a single channel - numpy array
    :param Fs: sampling frequency
    :return: x, data with artefacts set as nan
    """
    params = NEURAL_parameters.NEURAL_parameters()

    # DBverbose = 1
    N = len(x)  # verify / check this

    amount_removed = np.array([0, 0, 0, 0], dtype=float)
    # ---------------------------------------------------------------------
    # 1. electrode - checks(continuous row of zeros)
    # ---------------------------------------------------------------------
    x_channel = x.copy()
    x_channel[np.where(x_channel != 0)] = 1
    irem = np.zeros([N])
    lens, istart, iend = utils.len_cont_zeros(x_channel, 0)
    ielec = np.array(np.where(lens >= (params["ART_ELEC_CHECK"] * Fs)))[0]

    if ielec.size != 0:
        for m in ielec:
            irun = np.array(
                range((istart[m] - 1), (iend[m] + 2), 1)
            )  # TODO check this against matlab
            irun[np.where(irun < 0)] = 0
            irun[np.where(irun > N - 1)] = N - 1
            irem[irun] = 1
            x[irun] = np.nan
    if any(irem == 1):
        print("continuous row of zeros: %.2f%%\n" % (100 * (np.sum(irem) / N)))
    amount_removed[0] = 100 * (np.sum(irem) / N)

    x_nofilt = x.copy()
    x_filt, inans = utils.filter_butterworth_withnans(x, Fs, 40, 0.1, [5, 2])

    # ---------------------------------------------------------------------
    # 2. high - amplitude artefacts
    # ---------------------------------------------------------------------
    art_coll = params["ART_TIME_COLLAR"] * Fs
    irem = np.zeros(N)

    x_hilbert = np.abs(signal.hilbert(x_filt))

    thres_upper = params["ART_HIGH_VOLT"]
    ihigh = np.array(np.where(x_hilbert > thres_upper))[0]

    if ihigh.size != 0:
        for p in range(len(ihigh)):
            irun = np.array(
                range((ihigh[p] - int(art_coll)), (ihigh[p] + int(art_coll) + 1), 1)
            )
            irun[np.where(irun < 0)] = 0
            irun[np.where(irun > N - 1)] = N - 1
            irem[irun] = 1
    x[irem == 1] = np.nan
    if any(irem == 1) and DBverbose:
        print(
            "length of high-amplitude artefacts: %.2f%%\n" % (100 * (np.sum(irem) / N))
        )
    amount_removed[1] = 100 * (np.sum(irem) / N)

    # ---------------------------------------------------------------------
    # 3. continuous constant values(i.e.artefacts)
    # ---------------------------------------------------------------------
    art_coll = params["ART_DIFF_TIME_COLLAR"] * Fs
    irem = np.zeros(N)

    x_diff_all = np.concatenate((np.diff(x), [0]))
    x_diff = x_diff_all.copy()
    x_diff[x_diff != 0] = 1
    lens, istart, iend = utils.len_cont_zeros(x_diff, 0)

    # if exactly constant for longer than.then remove:
    ielec = np.array(np.where(lens > (params["ART_DIFF_MIN_TIME"] * Fs)))[0]

    if ielec.size != 0:
        for m in ielec:
            irun = np.array(
                range((istart[m] - int(art_coll)), (iend[m] + int(art_coll) + 1), 1)
            )
            irun[np.where(irun < 0)] = 0
            irun[np.where(irun > N - 1)] = N - 1
            irem[irun] = 1
            x[irun] = np.nan
    if any(irem == 1):
        print(
            "continuous row of constant values: %.2f%%\n" % (100 * (np.sum(irem) / N))
        )
    amount_removed[2] = 100 * (np.sum(irem) / N)

    # ---------------------------------------------------------------------
    # 4. sudden jumps in amplitudes or constant values(i.e.artefacts)
    # ---------------------------------------------------------------------
    art_coll = params["ART_DIFF_TIME_COLLAR"] * Fs
    irem = np.zeros(N)
    x_diff = x_diff_all.copy()

    ihigh = np.array(np.where(np.abs(x_diff) > params["ART_DIFF_VOLT"]))[0]
    if ihigh.size != 0:
        for p in range(len(ihigh)):
            irun = np.array(
                range((ihigh[p] - int(art_coll)), (ihigh[p] + int(art_coll) + 1), 1)
            )
            irun[np.where(irun < 0)] = 0
            irun[np.where(irun > N - 1)] = N - 1
            irem[irun] = 1

    x[irem == 1] = np.nan
    if any(irem == 1) and DBverbose:
        print("length of sudden-jump artefacts: %.2f%%\n" % (100 * (np.sum(irem) / N)))
    amount_removed[3] = 100 * (np.sum(irem) / N)

    # before filtering, but should be eliminated anyway
    x[inans] = np.nan
    inans = np.where(np.isnan(x))
    x_nofilt[inans] = np.nan
    # x_nofilt[irem_muc == 1] = np.inf
    x = x_nofilt

    return x, amount_removed


def remove_artefacts(data, ch_labels, Fs, data_ref, ch_labels_ref):
    """
    remove_artefacts: simple procedure to remove artefacts

    Syntax: data = remove_artefacts(data, ch_labels, Fs, data_ref, ch_refs)

    Inputs:
        data      - EEG data, in bipolar montage; size: N_channels x N (pandas dataframe)
        ch_labels - List of bipolar channel labels,
                    e.g. ['C3-O1','C4-O2', 'F3-C3', 'F4-C4']
        Fs        - sampling frequency (in Hz)
        data_ref  - EEG data, in referential  montage; size: (N_channels+1) x N
        ch_refs   - List of referential channel labels,
                    e.g. ['C3','C4','F3','F4']

    Outputs:
        data - EEG data after processing, in bipolar montage, size: N_channels x N  (pandas dataframe)

    Example:
        import utils
        import numpy as np
        import preprocessing_EEG
        Fs = 256
        data_st = utils.gen_test_EEGdata(2*60, Fs, 1)
        N = len(data_st['eeg_data_ref'])

        # simulate artefacts:
        # 1. F3 not properly attached:
        if3 = [i for i, x in enumerate(data_st['ch_labels_ref']) if x == 'F3'][0]
        data_st['eeg_data_ref'][data_st['ch_labels_ref'][if3]] = np.random.randn(N) * 10

        # 2. electrode coupling between C4 and Cz
        ic4 = [i for i, x in enumerate(data_st['ch_labels_ref']) if x == 'C4'][0]
        icz = [i for i, x in enumerate(data_st['ch_labels_ref']) if x == 'Cz'][0]

        data_st['eeg_data_ref'][data_st['ch_labels_ref'][icz]] = data_st['eeg_data_ref'][data_st['ch_labels_ref'][ic4]]
                                                                + np.random.randn(N) * 5

        # re-generate bipolar montage:
        data_st['eeg_data'], data_st['ch_labels'] = utils.set_bi_montage(data_st['eeg_data_ref'], Fs,
                                                    data_st['ch_labels_bi'])

        # remove channels:
        eeg_art = preprocessing_EEG.remove_artefacts(data_st['eeg_data'].copy(), data_st['ch_labels'], data_st['Fs'],
                                         data_st['eeg_data_ref'], data_st['ch_labels_ref'])
    """
    params = NEURAL_parameters.NEURAL_parameters()
    DBverbose = 1

    N, N_channels = data.shape

    # ---------------------------------------------------------------------
    # 0. check in referential mode first; is there problem with one
    #    channel(e.g.Cz)
    # ---------------------------------------------------------------------
    irem_channel = np.array([], dtype=int)
    channel_names_ref = list(data_ref.columns)
    if not data_ref.empty:
        x_filt = np.zeros([data_ref.shape[1], data_ref.shape[0]])
        for ch in range(data_ref.shape[1]):
            x_filt[ch, :], dum = utils.filter_butterworth_withnans(
                data_ref[channel_names_ref[ch]].to_numpy(), Fs, 20, 0.5, 5
            )

        r = np.corrcoef(x_filt)
        np.fill_diagonal(r, np.nan)
        r_channel = np.nanmean(r, axis=0)
        del x_filt

        ilow = np.array(
            np.where(np.abs(r_channel) < params["ART_REF_LOW_CORR"])
        ).astype(int)[0]

        if ilow.size != 0:
            nn = 1
            irem_channel = np.array([], dtype=int)
            for idex in ilow:
                ch_find = ch_labels_ref[idex]
                itmp = [i for i, x in enumerate(ch_labels) if ch_find in x]
                irem_channel = np.append(irem_channel, itmp)
                nn += 1
            print(
                ":: remove channel (low ref. correlation): %s\n"
                % np.array(ch_labels)[irem_channel.astype(int)]
            )

        # if DBverbose:
        # print(r_channel)
        # print(ch_labels_ref)
        for ch in range(len(irem_channel)):
            data[ch_labels[irem_channel[ch].astype(int)]] = np.nan

    ichannels = list(range(N_channels))
    ichannels = np.delete(ichannels, irem_channel)
    N_channels = len(ichannels)

    # ---------------------------------------------------------------------
    # 1. look for electrode coupling:
    # ---------------------------------------------------------------------
    ch_labels = np.array(ch_labels)
    if N_channels > 4:
        ileft, iright = channel_hemispheres(ch_labels[ichannels])

        if len(ileft) > 1 and len(iright) > 1:
            x_means = np.zeros(N_channels)

            x_filt = np.zeros([data_ref.shape[1], data_ref.shape[0]])
            for ch in range(N_channels):
                x_filt[ch, :], dum = utils.filter_butterworth_withnans(
                    data[ch_labels[ichannels[ch]]].to_numpy(), Fs, 20, 0.5, 5
                )

            A = []  # to suppress warning
            for n in range(N_channels):
                x_means[n] = np.nanmean(np.abs(x_filt[n, :]) ** 2)

                if n == 0:
                    A = [[x_means[n], ch_labels[ichannels[n]]]]
                else:
                    A.append([x_means[n], ch_labels[ichannels[n]]])
            del x_filt

            # 1/4 of the median channel energy
            cut_off_left = np.median(x_means[ileft]) / 4
            cut_off_right = np.median(x_means[iright]) / 4

            ishort_left = np.array(np.where(x_means[ileft] < cut_off_left))
            ishort_right = np.array(np.where(x_means[iright] < cut_off_right))

            ishort = np.hstack([ileft[ishort_left], iright[ishort_right]])
            ishort = np.array(ichannels[ishort])

            if ishort.size != 0:
                print(
                    ":: remove channel (electrode coupling): %s\n" % ch_labels[ishort]
                )
                data[ch_labels[ishort][0]] = np.nan
                irem_channel = np.append(irem_channel, ishort[0])

    # if DBverbose and len(A) > 0:
    #    print(A)

    ichannels = list(range(len(data.columns)))
    ichannels = np.delete(ichannels, irem_channel)

    # all other artefacts are on a channel-by-channel basis

    irem = np.array([])

    amount_removed = np.zeros([1, 4])
    ct = 0

    for n in ichannels:
        data[ch_labels[n]], tmpp = art_per_channel(
            data[ch_labels[n]].to_numpy(), Fs, DBverbose
        )
        if ct == 0:
            amount_removed[ct, :] = tmpp
        else:
            amount_removed = np.vstack((amount_removed, tmpp))
        ct += 1
        irem = np.append(irem, np.where(np.isnan(data[ch_labels[n]])))
        if any(np.isinf(data[ch_labels[n]])):
            data[ch_labels[n]][np.where(data[ch_labels[n]] == np.inf)[0]] = np.nan

    # remove artefacts across all channels

    data.loc[np.unique(irem), ch_labels[ichannels]] = np.nan

    out = list(np.max(amount_removed, axis=0))
    if irem_channel.shape[0] != 0:
        out.append(np.unique(ch_labels[irem_channel.astype(int)]))
    else:
        out.append("None removed")

    return data, out


def LPF_zero_phase(pd_data, Fs):
    """

    (pd) -> pd

    Pass in pandas dataframe and low pass filter all the columns

    > LPF_zero_phase(data)
    data

    """
    params = NEURAL_parameters.NEURAL_parameters()
    lp = params["LP_fc"]

    cols = list(pd_data.columns)
    for i in range(len(cols)):
        data = pd_data[cols[i]].to_numpy()
        if all(np.isnan(data)):
            continue
        d_mean = np.nanmean(data)
        data = data - d_mean
        s1_filt = utils.filter_zerophase(data, Fs=Fs, LP_fc=lp, L_filt=int(4001))
        data = s1_filt + d_mean
        pd_data[cols[i]] = data

    return pd_data


def signal_downsample(pd_data, Fs, Fs_new):
    """

    (pd_data,int,int) -> pd_data

    The purpose of this function is to downsample pd_data from fs to fs_new

    > signal_downsample(pd_data,fs,fs_new)
    pd_data

    """

    cols = list(pd_data.columns)

    def isint(x):
        return x == round(x, 1)

    pd_data_downsample = pd.DataFrame()
    loop_index = 0

    if isint(Fs / Fs_new):
        idec = list(range(0, len(pd_data), round(Fs / Fs_new)))

        for i in range(len(cols)):
            data = np.array(pd_data[cols[i]])
            eeg_data = data[idec]
            pd_data_downsample.insert(loop_index, cols[i], eeg_data)
            loop_index += 1
    else:
        size_resampled = signal.resample_poly(
            np.ones(len(pd_data)), round(len(pd_data) / (Fs / Fs_new)), len(pd_data)
        ).size

        for i in range(
            len(cols)
        ):  # To find the size of resampled array so rows of all nans are sorted
            data = np.array(pd_data[cols[i]])
            if sum(np.isnan(data)) == len(data):
                down_eeg_data = np.ones(size_resampled)
                down_eeg_data.fill(np.nan)
                pd_data_downsample.insert(loop_index, cols[i], down_eeg_data)
                loop_index += 1
                continue

            down_eeg_data = signal.resample_poly(
                data, round(len(data) / (Fs / Fs_new)), len(data)
            )
            pd_data_downsample.insert(loop_index, cols[i], down_eeg_data)
            loop_index += 1

    return pd_data_downsample, Fs_new


def main_preprocessing(file, save_converted=None, save=0, Fs_new=64):
    """
    This function is used to  a) read in EEG from .edf files
                              b) remove artefacts
                              c) band-pass filter
                              d) downsample
                              e) save as .csv file

    """
    data, ch_fs = load_edf_EEG_file(file)
    Fs = ch_fs[0]  # Assuming all EEG channels

    montage_data, dum = utils.set_bi_montage(data, Fs)

    clean_data, amount_removed = remove_artefacts(
        montage_data.copy(),
        np.array(list(montage_data.columns)),
        Fs,
        data,
        np.array(list(data.columns)),
    )

    filtered_data = LPF_zero_phase(clean_data.copy(), Fs)

    down_sampled_data, Fs = signal_downsample(filtered_data.copy(), Fs, Fs_new)

    packaged_data = {
        "eeg_data": down_sampled_data,
        "Fs": Fs,
        "ch_labels": list(down_sampled_data.columns),
    }

    if save and not not save_converted:
        utils.save_pd_as_file(down_sampled_data, save_converted + "down", Fs)

    return packaged_data, amount_removed
